# -*- coding: utf-8 -*-
"""
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
conditioning.py 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module includes the routines used to precondition the input data
before any extraction is done like interpolating the
data points, normalizing the data and filtering points that are too noisy.

Example
-------
In order to normalize data the following steps have to be executed either in a script or
python3 terminal::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)
	norm_value = 35.8/10**9
	fompy.normalize(fds, norm_value)

In order to filter data the following steps have to be executed either in a script or
python3 terminal::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)
	fompy.filter(fds, theta_crit = 1.52)

Additionally, by default FoMpy uses the cubic spline interpolation in order to 
extract accurately the FoMs. If the user wishes to change the interpolation methods, 
there are several implemented and tested methods that can be selected. 
The list of available methods includes: 'akima','pchip' or 'linear'::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)
	fds.interpolation = 'linear'

"""


from abc import ABCMeta, abstractmethod
import numpy as np
#----------------------------------------------------------------------------------------------------------------
class _normalizerStrategy(metaclass=ABCMeta):
	"""
	Abstract class for normalizing a semiconductor's IV curve.

	"""

	@abstractmethod
	def normalize(self):
		pass

#----------------------------------------------------------------------------------------------------------------
class normalizer(_normalizerStrategy):
	"""
	Normalizer class.

	...

	Methods
	-------
	normalize(fds, norm)
		Class method used for normalizing the currents a semiconductor's IV curve.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	norm : float
		float value used to normalize the currents contained in the FoMpy dataset IV curves.
	"""

	def normalize(self, fds, norm):
		try:
			for i in range(len(fds.dataset)):
				fds.dataset[i][:,1][:] = np.divide(fds.dataset[i][:,1][:],norm)
		except TypeError:
			pass
#----------------------------------------------------------------------------------------------------------------
class _interpolationStrategy(metaclass=ABCMeta):
	"""
	Abstract class containing several tested interpolation strategies that can be used in a semiconductor's IV curve.

	"""

	@abstractmethod
	def spline_interpol(self, x,y,s):
		pass

	@abstractmethod
	def akima_interpol(self,  x,y,s):
		pass        

	@abstractmethod
	def pchip_interpol(self,  x,y,s):
		pass  

	@abstractmethod
	def lin_interpol(self,x,y):
		pass  

#----------------------------------------------------------------------------------------------------------------
class interpolator(_interpolationStrategy):
	"""
	Class containing several tested interpolation strategies that can be used in a semiconductor's IV curve.
	For further documentation go to  https://docs.scipy.org/doc/scipy/reference/interpolate.html

	...

	Methods
	-------
	spline_interpol(x ,y , n , d , s)
		Interpolate data with a piecewise cubic polynomial which is twice
		continuously differentiable [1]. The result is represented as a PPoly
		instance with breakpoints matching the given data. The natural boundary
		condition is selected by default.

	Parameters
	----------
	x : array_like, shape (n,)
		1-d array containing values of the independent variable.
	y : array_like
		Array containing values of the dependent variable.
		It can have arbitrary number of dimensions, but the length along axis
		must match the length of x. Values must be finite.
	n : int
		Length of the output interpolated array.
	d : int
		Degree of the smoothing spline. Must be <= 5.
		Default is k=3, a cubic spline.
	s : float
		Positive smoothing factor used to choose the number of knots. 

	Methods
	-------
	akima_interpol(x,y,n)
		Fit piecewise cubic polynomials, given vectors x and y.
		The interpolation method by Akima uses a continuously differentiable
		sub-spline built from piecewise cubic polynomials. The resultant curve passes
		through the given data points and will appear smooth and natural.

	Parameters
	----------
	x : array_like, shape (n,)
		1-D array of monotonically increasing real values.
	y : array_like
		N-D array of real values. The length of y
		along the first axis must be equal to the length of x.
	n : int
		Length of the output interpolated array.

	Methods
	-------
	pchip_interpol(x, y, n)
		Convenience function for pchip interpolation. xi and yi are arrays of
		values used to approximate some function f, with yi = f(xi).
		The interpolant uses monotonic cubic splines to find the value of
		new points x and the derivatives there.

	Parameters
	----------
	x : array_like, shape (n,)
		A 1-D array of monotonically increasing real values.
		x cannot include duplicate values (otherwise f is overspecified)
	y : array_like
		A 1-D array of real values. y’s length along
		the interpolation axis must be equal to the length
		of x. If N-D array, use axis parameter to select correct axis.
	n : int
		Length of the output interpolated array.

	Methods
	-------
	lin_interpol(x,y, n)
		Interpolate a 1-D function.x and y are arrays of values used to approximate
		some function f: y = f(x). This class returns a function
		whose call method uses interpolation to find the value of new points.

	Parameters
	----------
	x : array_like, shape (n,)
		A 1-D array of real values.
	y : array_like
		A N-D array of real values. The length of y along
		the interpolation axis must be equal to the length of x.
	n : int
		Length of the output interpolated array.

	"""
	def spline_interpol(self, x = None ,y = None, n = None, d = None, s = 0):

		if(d is 3) or (d is None):
			from scipy.interpolate import CubicSpline
			cs_natural = CubicSpline(x, y, bc_type='natural')
			xvals = np.linspace(x[0], x[-1], n)
			return (xvals, cs_natural(xvals))
		elif(d is 5) :
			from scipy import interpolate
			spline = interpolate.splrep(x, y, k=5)
			xvals = np.linspace(x[0], x[-1], n)
			yinterp = interpolate.splev(xvals, spline, der=0)
			return (xvals, yinterp)
		else:
			pass

	def akima_interpol(self, x,y,n):
		from scipy.interpolate import Akima1DInterpolator
		cs_natural = Akima1DInterpolator(x, y, axis=0)
		xvals = np.linspace(x[0], x[-1], n)
		return (xvals, cs_natural(xvals))

	def pchip_interpol(self, x, y, n):
		from scipy.interpolate import PchipInterpolator
		cs_natural = PchipInterpolator(x, y)
		xvals = np.linspace(x[0], x[-1], n)		
		return (xvals, cs_natural(xvals)) 

	def lin_interpol(self,x,y, n):
		from scipy.interpolate import interp1d
		cs_natural = interp1d(x, y)
		xvals = np.linspace(x[0], x[-1], n)		
		return (xvals, cs_natural(xvals))

#----------------------------------------------------------------------------------------------------------------
class _filterStrategy(metaclass=ABCMeta):
	"""
	Abstract class for filtering data of a semiconductor's IV curve.

	"""

	@abstractmethod
	def polar_filter(self):
		pass

	# @abstractmethod
	# def gauss_filter(self):
	# 	pass

#--------------------------------------------------------------------------------------------------------------
class filter_tool(_filterStrategy):
	"""
	Class for filtering noisy data from a semiconductor's IV curve.

	...

	Methods
	-------
	polar_filter(fds,theta_crit, show_theta = False)
		Class method that filters data from a semiconductor's IV curve checking the increase in the angle
		of the points with respect to the origin of an IV curve and removes all the increments that go above
		defined a threshold.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	theta_crit : float
		Threshold value for the biggest allowed increase in the angle between to consecutive points
		of an IV curve.
	show_theta : bool
		If True it prints all the angle increments between two consecutive points so the user can
		choose a suitable theta crit.	
	"""

	def polar_filter(self,fds,theta_crit, show_theta = False):
		for i in range(len(fds.dataset)):
			try:
				theta = angle_wrt_0(fds.dataset[i][:,0],np.log10(fds.dataset[i][:,1])) #The angle wrt 0 is calculated
				step = np.gradient(angle_wrt_0(fds.dataset[i][:,0],np.log10(fds.dataset[i][:,1]))) #The difference between the angles is calculated
				
				j = 0
				step_lst = list(np.abs(step))
				if(show_theta is True):
					print(step_lst)
				indexes = []
				for index, elem in list(enumerate(step_lst)):
					if elem > theta_crit: #If the angle increment is above threshold the index is stored in 'indexes'
						indexes.append(index-j)
						step_lst.pop(index-j)
						j = j+1
			except (TypeError, ValueError):
				pass

			for item in indexes: #And then deleted
				fds.dataset[i] = np.delete(fds.dataset[i],item, axis=0)


#----------------------------------------------------------------------------------------------------------------
def angle_wrt_0(x, y):
	"""Function that calculates the polar angle of a point with respect to the origin of coordinates.
	It returns the angle in degrees and it is assummed that the angle is contained in the 1rst cuadrant.

	Parameters
	----------
	x: float
		The x-axis coodinate of the point.
	y: float
		The y-axis coodinate of the point.

	Returns
	-------
	theta: float
		Angle between the given point and the origin
		of coordinates in degrees. It is assummed that the angle is contained
		in the 1rst cuadrant.

	"""
	if(len(x) is not len(y)):
		raise('The sizes of x and y must be the same')
	if(len(x) is 1):
		if(x == 0):
			theta = 90
		else:
			theta = np.arctan(y/x)*180/np.pi
	else:
		theta = []
		for i in range(len(x)):
			if(x[i] == 0):
				theta.append(90)
			else:
				theta.append(np.arctan(y[i]/x[i])*180/np.pi)
	return np.abs(theta) #Theta is considered to be in the first cuadrant



from __future__ import division, absolute_import, print_function

from .fds import *
from .fom import *
from .plots import *
from .conditioning import *
from .aux import *
from .wrappers import *

name = "mlfompy"
import os

__version__ = '0.0.1'






# -*- coding: utf-8 -*-
"""
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
wrappers.py 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module includes all the wrapper functions used in the FoMpy library.
Code examples on how to use them can be seen either in the complete guide, the repository quickstart
or at the beginning of each file of the source code.

"""

from fompy.fds import daoFile
from fompy.fom import vth_ext, ioff_ext, ion_ext, ss_ext, dibl_ext
from fompy.conditioning import normalizer, filter_tool
from fompy.plots import plotter
from fompy.aux import checkPath
from glob import glob
import os, re

def dataset(path = None,arr = None,  parser = None, save_to_file = None, interval = None, exclude = None):
	"""
	Wrapper function that creates a FoMpy dataset.

	Parameters
	----------
	path : str
		Path to the file containing the IV curves
	arr : array_like
		Array of data containing one or more semiconductor's IV curves.
	parser : void
		Format to read the file
	save_to_file : str
		Path of the file to store all the FoMpy dataset
	interval : array_like
		List of two int values: start(index of the first simulation to load into the FompyDataset)
		and end(index of the last simulation to load into the FompyDataset)
	exclude : array_like
		Index values of simulations to exclude.

	Returns
	-------
	FompyDataset
		Class containing the most important parameters of a semiconductor IV curve

	"""
	dao_dataset = daoFile()
	fds = dao_dataset.load(path = path,arr = arr , parser = parser , interval = interval, exclude = exclude)
	if(save_to_file != None):
		dao_dataset.save(fds, save_to_file)
	return fds
def iv(path = None,arr = None,  parser = None, save_to_file = None):
	"""
	Wrapper function that creates a FoMpy dataset from a single IV curve.

	Parameters
	----------
	path : str
		Path to the file containing the IV curves
	arr : array_like
		Single IV curve.
	parser : void
		Format to read the file
	save_to_file : str
		Path of the file to store all the FoMpy dataset

	Returns
	-------
	FompyDataset
		Class containing the most important parameters of a semiconductor IV curve

	"""
	dao_dataset = daoFile()
	iv = arr
	fds = dao_dataset.load(path = path, parser = parser, iv = iv)
	if(save_to_file != None):
		dao_dataset.save(fds, save_to_file)
	return fds
#----------------------------------------------------------------------------------------------------------------
def extract(fds1, fds2 = None, fom = None, method = None, cc_criteria = None, vg_ext = None, print_fom = None,
 vg_start = None, vg_end = None):
	"""
	Wrapper function that extracts the most common figures of merit from a FoMpy dataset.

	Parameters
	----------
	fds1 : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
		Needed for the extraction of any FoM.
	fds2 : FoMpy Dataset
		additional structure of data containing the most important parameters of a semiconductor's IV curve.
		Needed for the extraction of the DIBL.
	fom : str
		Keyword indicating the desired FoMs to extract. The list of available FoMs includes:
		'vth', 'ioff', 'ion', 'ss', 'ratio', 'power' and 'dibl'.
	method : str
		Keyword indicating the desired method of extraction of the FoMs.The list of available methods includes:
		'SD', 'CC', 'TD' and 'LE'. If method is not defined the 'SD' is selected by default.
	cc_criteria : float, optional
		Float value used for the extraction of several FoMs using the constant current method.
	vg_ext : float
		Gate voltage value used to calculate a FoM like IOFF or ION.
	print_fom : bool
		If True all the FoMs are extracted from a FoMpy Dataset (except for the DIBL)
	vg_start : float
		Gate voltage defining the start of the interval in which the Subthreshold Swing is extracted.
	vg_end : float
		Gate voltage defining the end of the interval in which the Subthreshold Swing is extracted.

	Returns
	-------
	parameter : array_like
		1-d array containing the extracted FoMs.

	""" 
	if(str(fom) == 'vth'):
		temp_vth = vth_ext()
		if (method != 'LE'):
			parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
		else:
			parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)			
		return parameter_vth
	elif(str(fom) == 'ioff'):
		temp_ioff = ioff_ext()
		parameter_ioff, curves= temp_ioff.extraction(fds1,vg_ext = vg_ext)
		return parameter_ioff
	elif(str(fom) == 'ion'):
		temp_ion = ion_ext()
		if(vg_ext != None) and (type(vg_ext) == float):
			parameter_ion, curves= temp_ion.extraction(fds1,vg_ext = vg_ext)
		else:
			temp_vth = vth_ext()
			if (method != 'LE'):
				parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
			else:
				parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)	
			parameter_ion, curves= temp_ion.extraction(fds1,vth = parameter_vth)
		return parameter_ion
	elif(str(fom) == 'ss'):
		temp_ss = ss_ext()
		if(vg_end == None):
			if(vg_start == None):
				vg_start = 0.0
			temp_vth = vth_ext()
			if (method != 'LE'):
				parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
				parameter_ss, curves, _ = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
			else:
				parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
				parameter_ss, curves, _ = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
		elif(vg_end != None):
			if(vg_start == None):
				vg_start = 0.0			
			temp_vth = vth_ext()
			if (method != 'LE'):
				parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
				parameter_ss, curves, _ = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
			else:
				parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
				parameter_ss, curves, _ = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
		else:
			pass
			

		return parameter_ss
	elif(fom == 'dibl'):
		temp = dibl_ext()
		parameter_dibl,curve_high, curve_low, vth_high, vth_low, corriente_low = temp.extraction(fds1,fds2, method, cc_criteria)
		return parameter_dibl

	elif(print_fom == True):

		temp_vth = vth_ext()
		if (method != 'LE'):
			parameter_vth_1, curves= temp_vth.extraction(fds1, method, cc_criteria)
			if(fds2 != None):
				parameter_vth_2, curves= temp_vth.extraction(fds2, method, cc_criteria)

		else:
			parameter_vth_1, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
			if(fds2 != None):
				parameter_vth_2, curves, A, B= temp_vth.extraction(fds2, method, cc_criteria)
		
		temp_ioff = ioff_ext()
		parameter_ioff_1, curves= temp_ioff.extraction(fds1,vg_ext = vg_ext)
		if(fds2 != None):
			parameter_ioff_2, curves= temp_ioff.extraction(fds2,vg_ext = vg_ext)

		temp_ion = ion_ext()
		if(vg_ext != None) and (type(vg_ext) is float):
			parameter_ion_1, curves= temp_ion.extraction(fds1,vg_ext = vg_ext)
			if(fds2 != None):
				parameter_ion_2, curves= temp_ion.extraction(fds2,vg_ext = vg_ext)
		else:
			temp_vth = vth_ext()
			if (method != 'LE'):
				parameter_vth_1, curves= temp_vth.extraction(fds1, method, cc_criteria)
				if(fds2 != None):
					parameter_vth_2, curves= temp_vth.extraction(fds2, method, cc_criteria)

			else:
				parameter_vth_1, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
				if(fds2 != None):
					parameter_vth_2, curves, A, B= temp_vth.extraction(fds2, method, cc_criteria)

			parameter_ion_1, curves= temp_ion.extraction(fds1,vth = parameter_vth_1)
			if(fds2 != None):
				parameter_ion_2, curves= temp_ion.extraction(fds2,vth = parameter_vth_2)

		temp_ss = ss_ext()
		parameter_ss_1, curves, _ = temp_ss.extraction(fds1, parameter_vth_1, vg_start, vg_end)
		if(fds2 != None):
			parameter_ss_2, curves, _ = temp_ss.extraction(fds2, parameter_vth_2, vg_start, vg_end)

		if((fds1 != None) and (fds2 != None)):
			temp = dibl_ext()
			parameter_dibl,curve_high, curve_low, vth_high, vth_low, corriente_low = temp.extraction(fds1,fds2, method, cc_criteria)
			parameter_fds1 = (parameter_vth_1, parameter_ioff_1, parameter_ion_1, parameter_ss_1)
			parameter_fds2 = (parameter_vth_2, parameter_ioff_2, parameter_ion_2, parameter_ss_2)
			print('\nThe extracted values for fds1 are\nvth: {0}\nioff: {1}\nion: {2} \nss: {3}\n'.format(parameter_vth_1, parameter_ioff_1, parameter_ion_1, parameter_ss_1) )
			print('The extracted values for fds2 are\nvth: {0}\nioff: {1}\nion: {2} \nss: {3}\n'.format(parameter_vth_2, parameter_ioff_2, parameter_ion_2, parameter_ss_2) )
			print('The extracted value for the DIBL are\n{0}\n'.format(parameter_dibl) )
			return (parameter_fds1,parameter_fds2)
		else:
			print('\nThe extracted values for fds1 are\nvth: {0}\nioff: {1}\nion: {2} \nss: {3}\n'.format(parameter_vth_1, parameter_ioff_1, parameter_ion_1, parameter_ss_1) )
			return (parameter_vth_1, parameter_ioff_1, parameter_ion_1, parameter_ss_1)
	else:
		raise Exception('The figure of merit hasn\'t been defined!')
#----------------------------------------------------------------------------------------------------------------
def plot(fds1, fds2 = None, plot_type = None, fom = None,var_source= None, parameter=None,  method = None,bins = None, cc_criteria = None, vg_ext = None, vg_start = None, vg_end = None,cont_parameter = None,  backend = None, save_plot = None):
	"""
	Wrapper function that plots the most common figures in semiconductor simulations.

	Parameters
	----------
	fds1 : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
		Needed for generating the plot of any FoM.
	fds2 : FoMpy Dataset
		additional structure of data containing the most important parameters of a semiconductor's IV curve.
		Needed for generating the plot of the calibration and the DIBL.
	plot_type : str
		Keyword indicating the type of plot to generate. The list of available plots includes:
		'iv', 'hist', 'qq', 'varplot', 'calib' and 'fomplot'.
	fom : str
		Keyword indicating the FoM to be plotted. The list of available methods includes:
		'vth', 'ioff', 'ion', 'ss', 'ratio', 'power' and 'dibl'.
	parameter : array_like
		Array of extracted FoM values to be plotted.
	method : str
		Keyword indicating the desired method of extraction of the FoMs. The list of available methods includes:
		'SD'(default), 'CC', 'TD' and 'LE'. If method is not defined the 'SD' is selected by default.
	bins : int
		It defines the number of equal-width bins in the given range.
	cc_criteria : float
		Current criteria used to extract vth with the CC criteria for the fomplot.
	var_source : str
		Type of deformations applied to study varibiality. For example, for metal grain granularity the grain size.
	vg_ext : float
		Gate voltage value used to calculate a FoM like IOFF or ION.
	vg_start : float
		Gate voltage defining the start of the interval in which the Subthreshold Swing is extracted.
	vg_end : float
		Gate voltage defining the end of the interval in which the Subthreshold Swing is extracted.
	backend : str
		String containing the name of the backend chosen to either plot or save the plots. The backends available are:
		'Agg'(default), which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tool for visualizing the plots on a pop-up window.
		'TkAgg' requires the package python3-tk installed in order to run.
	save_plot : bool
		If True the generated plot is save to the defined path.

	""" 

	plot = plotter()
	if(str(plot_type) == 'iv'):
		plot.iv(fds1,backend = backend, save_plot=save_plot)
	elif(str(plot_type) =='hist'):
		plot.hist(parameter=parameter,bins=bins,var_source = var_source, backend = backend,cont_parameter=cont_parameter, save_plot=save_plot)			
	elif(str(plot_type) == 'qq'):
		plot.qq(var_source =var_source, parameter=parameter, backend = backend, save_plot=save_plot,method=None)	
	elif(str(plot_type) == 'varplot'):
		plot.varplot(fds1,var_source = var_source, backend = backend, save_plot=save_plot)
	elif(str(plot_type) == 'varplotlogCC'):
		plot.varplotlogCC(fds1,cc_criteria = cc_criteria, var_source = var_source, backend = backend, save_plot=save_plot)	
	elif(str(plot_type) == 'calib'):
		plot = plotter()
		plot.calib(fds1,fds2,backend = backend, save_plot=save_plot)	
	elif(plot_type == None):
		if(str(fom) == 'vth'):
			temp_vth = vth_ext()
			if (method != 'LE'):
				parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
				temp_vth.plot(fds1 = fds1, parameter = parameter_vth,method=method,cc_criteria=cc_criteria, curves = curves,backend = backend, save_plot = save_plot)
			else:
				parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
				temp_vth.plot(fds1 = fds1, parameter = parameter_vth,method=method,cc_criteria=cc_criteria, curves = curves,backend = backend, save_plot = save_plot, A =A, B=B)
		elif(str(fom) == 'ioff'):
			temp_ioff = ioff_ext()
			parameter_ioff, curves= temp_ioff.extraction(fds1,vg_ext)
			temp_ioff.plot(fds1 = fds1, parameter = parameter_ioff, curves = curves,vg_ext = vg_ext,backend = backend, save_plot = save_plot)
		elif(str(fom) == 'ion'):
			temp_ion = ion_ext()
			if(vg_ext != None) and (type(vg_ext) == float):
				parameter_ion, curves= temp_ion.extraction(fds1,vg_ext = vg_ext)
				print(parameter_ion)
				temp_ion.plot(fds1 = fds1, parameter = parameter_ion, curves = curves,vg_ext = vg_ext,backend = backend, save_plot = save_plot)
			else:
				temp_vth = vth_ext()
				if (method != 'LE'):
					parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
				else:
					parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)	
				parameter_ion, curves= temp_ion.extraction(fds1,vth = parameter_vth)
				temp_ion.plot(fds1 = fds1, parameter = parameter_ion, curves = curves,parameter_vth=parameter_vth,backend = backend, save_plot = save_plot)
		elif(str(fom) == 'ss'):
			temp_ss = ss_ext()

			if(vg_end == None):
				if(vg_start == None):
					vg_start = 0.0
				temp_vth = vth_ext()
				if (method != 'LE'):
					parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
					parameter_ss, curves, vt_sd_medio = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
				else:
					parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
					parameter_ss, curves, vt_sd_medio = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
			elif(vg_end != None):
				if(vg_start == None):
					vg_start = 0.0			
				temp_vth = vth_ext()
				if (method != 'LE'):
					parameter_vth, curves= temp_vth.extraction(fds1, method, cc_criteria)
					parameter_ss, curves, vt_sd_medio = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
				else:
					parameter_vth, curves, A, B= temp_vth.extraction(fds1, method, cc_criteria)
					parameter_ss, curves, vt_sd_medio = temp_ss.extraction(fds1,parameter_vth, vg_start = vg_start, vg_end = vg_end)
			else:
				pass
			temp_ss.plot(fds1 = fds1, parameter = parameter_ss, curves = curves, vg_start = vg_start, vg_end = vg_end, vt_sd_medio = vt_sd_medio,backend = backend, save_plot = save_plot)
		elif(str(fom) == 'dibl'):
			temp = dibl_ext()
			parameter_dibl,curve_high, curve_low, vth_high, vth_low, corriente_low  = temp.extraction(fds1,fds2, method)
			temp.plot(fds1, curve_high, curve_low, vth_high, vth_low, corriente_low,backend = backend,  save_plot = save_plot)
#----------------------------------------------------------------------------------------------------------------
def savetotxt(path,fom,parameter,fds):
	"""
	Wrapper function that saves to a text file the extracted FoMs.

	Parameters
	----------
	path : str
		Defines the path where the extracted results are saved to a file.
	fom : str
		Keyword indicating the FoM to be plotted. The list of available methods includes:
		'vth', 'ioff', 'ion', 'ss', 'ratio', 'power' and 'dibl'.
	parameter : array_like
		Array of extracted FoM values to be plotted.

	""" 	
	if(str(fom) == 'vth'):
		temp_vth = vth_ext()
		temp_vth.save_results_to_file(path, parameter,fds)
	elif(str(fom) == 'ioff'):
		temp_ioff = ioff_ext()
		temp_ioff.save_results_to_file(path, parameter,fds)
	elif(str(fom) == 'ion'):
		temp_ion = ion_ext()
		temp_ion.save_results_to_file(path, parameter,fds)
	elif(str(fom) == 'ss'):
		temp_ss = ss_ext()
		temp_ss.save_results_to_file(path, parameter,fds)
	elif(str(fom) == 'dibl'):
		temp_dibl = dibl_ext()
		temp_dibl.save_results_to_file(path, parameter,fds)
#----------------------------------------------------------------------------------------------------------------
def normalize(fds, norm):
	"""
	Wrapper function that normalizes the currents a semiconductor's IV curve.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	norm : float
		float value used to normalize the currents contained in the FoMpy dataset IV curves.

	""" 	
	fds.norm = norm
	temp_norm = normalizer()
	temp_norm.normalize(fds, norm)
#----------------------------------------------------------------------------------------------------------------
def filter(fds1, theta_crit, show_theta = False):
	"""
	Wrapper class for filtering noisy data from a semiconductor's IV curve.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	theta_crit : float
		Threshold value for the biggest allowed increase in the angle between to consecutive points
		of an IV curve.
	show_theta : bool
		If True it prints all the angle increments between two consecutive points so the user can
		choose a suitable theta crit.
	
	"""

	temp_filter = filter_tool()
	fds1.dataset = temp_filter.polar_filter(fds1, theta_crit=theta_crit, show_theta=show_theta)
#----------------------------------------------------------------------------------------------------------------		
def version():
	"""
	Function that prints the current installed version of FoMpy

	"""

	here = os.path.abspath(os.path.dirname(__file__))

	def read(*parts):
		with open(os.path.join(here, *parts), 'r') as fp:
			return fp.read()

	def find_version(*file_paths):
		version_file = read(*file_paths)
		version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",version_file, re.M)
		if version_match:
			return version_match.group(1)
		raise RuntimeError("Unable to find version string.")
	version=find_version("fompy", "__init__.py")

	print('FoMpy version: '+version)







# -*- coding: utf-8 -*-
"""
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
plots.py 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module includes the routines used to generate several common figures
in semiconductor simulations.

Example
-------

After the FoMs have been extracted, FoMpy includes several useful visualizing routines, 
commonly used in semiconductor simulations. Code examples explaining how to use them can be seen below.

In order to generate a plot of the FoMs (fomplot)::

	import fompy
	path_file_high = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)
	vth_array = fompy.extract(fds, fom = 'vth')
	fompy.plot(fds, fom = 'vth', save_plot='./vth_plots/sd/')

and FoMpy will generate a single graph for each simulation showing the IV curve with its correspondent FoM extraction criteria and the extracted FoM. Most of the FoMs are plotted the same way, except for the DIBL. This is because, if we want to calculate the DIBL
two FoMpy Datasets are needed::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	path_file_low = './data/sim_FinFET_vd_low/'

	fds_hdb = fompy.dataset(path_file_JCJB, parser='JCJB')
	fds_ldb = fompy.dataset(path_file_low, parser='JCJB')
	fds_hdb.drain_bias_value = 0.7
	fds_ldb.drain_bias_value = 0.05

	fompy.plot(fds_hdb, fds_ldb, fom = 'dibl', save_plot='./dibl/')

Additionally, other types of plots can be generated. The list of available plots includes: 'iv', 'hist', 'qq', 'varplot', 'calib' and 'fomplot'. These are some examples::

	path_file_var = './data/simulations/'
	fds_var = fompy.dataset(path_file_var, parser=fompy.JCJB)
	vth_array = fompy.extract(fds_var, fom = 'vth')

	fompy.plot(fds, plot_type='hist', parameter=vth_array, bins=10,cont_parameter= 0.38, save_plot='./variability/')
	fompy.plot(fds, plot_type='qq', parameter=vth_array, save_plot='./variability/')

	path_file_var = './data/simulations/'
	fds_var = fompy.dataset(path_file_var, parser=fompy.JCJB)
	fompy.plot(fds, plot_type='varplot', save_plot='./variability/')


	path_file_JCJB = './data/sim_FinFET_vd_high/'
	path_file_low = './data/sim_FinFET_vd_low/'

	fds_hdb = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)
	fds_hdb.print_parameters()
	fds_ldb = fompy.dataset(path_file_low, parser=fompy.JCJB)
	fds_ldb.print_parameters()

	norm_value = 35.8/10**9
	fompy.normalize(fds_hdb, norm_value)
	fompy.normalize(fds_ldb, norm_value)

	fompy.plot(fds_hdb,fds_ldb, plot_type='calib', save_plot='./calibration/')
"""

from abc import ABCMeta, abstractmethod
import numpy as np
from scipy import stats
from enum import Enum
import os
import probscale
from fompy.aux import find_closest, get_diff, checkPath
from fompy.conditioning import interpolator
import matplotlib.pyplot as plt


class plotStrategy(metaclass=ABCMeta):
	"""
	Abstract class containing several commonly used plots for semiconductor simulations.

	"""

	# def __init__(self, **kwargs):
	# 	self.type = None

	@abstractmethod
	def iv(self):
		pass

	@abstractmethod
	def hist(self):
		pass

	@abstractmethod
	def qq(self):
		pass

	@abstractmethod
	def varplot(self):
		pass

	@abstractmethod
	def varplotlogCC(self):
		pass

	@abstractmethod
	def calib(self):
		pass

	@abstractmethod
	def fomplot(self):
		pass	
class plotter(plotStrategy):
	"""
	Class used for generating common plots in semiconductor simulations.
	For extensive documentation on how to modify this code go to https://matplotlib.org/tutorials/index.html
	"""

	def iv(self,fds,backend = None, save_plot = None):
		"""
		Methods
		-------
		iv(fds, save_plot = None)
			Class method that filters data from a semiconductor's IV curve using Gaussian filtering.

		Parameters
		----------
		fds : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
		save_plot : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.

		"""

		import matplotlib
		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass

		for i in range(len(fds.dataset)):
			try:
				plt.close()
				fig, ax1 = plt.subplots()
				ax2 = ax1.twinx()
				ax1.plot(fds.dataset[i][:,0][:],fds.dataset[i][:,1][:],lw =0.4,color='red', label='Log')
				ax2.plot(fds.dataset[i][:,0][:],fds.dataset[i][:,1][:],lw =0.4,color='green', label='Lin')
				ax1.set_yscale('log')
				ax1.set_ylabel(r'$\mathrm{I_{D}} $',labelpad=20)
				ax1.set_xlabel(r'$\mathrm{V_{G}} $',labelpad=20)
				ax1.set_title('IV curve plot -> Curve %s' % str(i))
				handles2, labels2 = ax1.get_legend_handles_labels()
				labels2, ids2 = np.unique(labels2, return_index=True)
				ax1.legend(handles2, labels2, loc='best')
				ax2.set_yscale('linear')
				ax2.yaxis.set_visible(False)
				handles, labels = ax2.get_legend_handles_labels()
				labels, ids = np.unique(labels, return_index=True)
				ax2.legend(handles, labels, loc='lower right')
				plt.tight_layout()
				if(save_plot != None):
					try:
						checkPath(save_plot)
						plt.savefig(save_plot+'iv_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
					except NameError:
						print('No path to save the plots has been chosen')
				else:
					print('No path has been defined\nTrying to use GUI backend...')
					plt.show()
				plt.close()
			except(TypeError, ValueError):
				pass
		return


	def hist(self, bins = None, parameter = None,cont_parameter = None, var_source = None, backend = None, save_plot = None):
		"""
		Methods
		-------
		hist(bins = None, parameter = None, save_to_file = None):
			Plot a histogram.The return value is a tuple (n, bins, patches) or ([n0, n1, ...],
			bins, [patches0, patches1,...]) if the input contains multiple data.

		Parameters
		----------
		bins : int
			If an integer is given, bins + 1 bin edges are calculated and returned, consistent with 
			numpy.histogram
		parameter : array_like, shape (n,)
			Input values, this takes either a single array or a sequence of arrays
			which are not required to be of the same length.
		var_source : string or None, optional
			Information of the variability applied to device
		save_plot : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.
		"""

		import matplotlib
		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass

		if(len(parameter)<2):
			raise Exception('Not enough data to plot a histogram')

		parameter = [x for x in parameter if x == x] # Remove all NaN values from the parameter list
		plt.close()
		xmin = np.min(parameter)
		xmax = np.max(parameter)
		x_range = np.linspace(xmin*0.8, xmax*1.2, len(parameter))
		m, s = stats.norm.fit(parameter) # get mean and standard deviation
		pdf_g = stats.norm.pdf(x_range, m, s) # now get theoretical probability density function in our interval
		max_pdf = np.max(pdf_g)
		pdf_g = pdf_g/max_pdf
		weights = np.ones_like(parameter)/float(len(parameter))
		if(s==0):
			print('The normal fit has not been plotted as sigma is zero')
		if(bins==None):
			bins = 5 # default number of bines if parameter not defined
		if(cont_parameter is not None):
			plt.axvline(cont_parameter, color='blue', linestyle='dashed', linewidth=2, label='Ideal case') # plot the extracter parameter value for the ideal case
		w, bins, patches = plt.hist(parameter,bins,weights=weights, facecolor='green')
		plt.plot(x_range, np.max(w)*pdf_g, color = 'r', label='Normal Fit') # plot the normal distribution fit
		plt.axvline(np.mean(parameter), color='r', linestyle='dashed', linewidth=2)
		plt.title('Distribution parameters: '+r'$\mu='+str(("%0.3f"%m))+r', \sigma$='+str(("%0.3f"%s))+'\n'  )
		plt.xlabel('$V_{th}$ [V]')
		plt.ylabel('Frecuency')
		plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
		plt.tight_layout()
		if(save_plot is not None):
			try:
				checkPath(save_plot)
				plt.savefig(save_plot+'hist_%s.pdf' %str(var_source), bbox_inches='tight', format='pdf', dpi=1200 )
			except NameError:
				print('No path to save the plots has been chosen')
		else:
			print('No path has been defined\nTrying to use GUI backend...')
			plt.show()
		plt.close()

		return

	def qq(self, parameter,backend = None, save_plot = None):
		"""
		Methods
		-------
		qq( parameter, save_to_file = None):
			Plot a Quantile plot.Plotting positions are converted into quantiles
			or Z-scores based on a probability distribution

		Parameters
		----------

		parameter : array_like, shape (n,)
			Input values, this takes either a single array or a sequence of arrays
			which are not required to be of the same length.
		save_plot : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.		

		"""

		import matplotlib
		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass

		if(len(parameter)<2):
			raise Exception('Not enough data for a QQ-plot')
			
		parameter = [x for x in parameter if x == x] # Remove all NaN values from the parameter list
		plt.close()
		fig, ax = plt.subplots()
		slope, r = probscale.probplot(parameter, ax=ax, plottype='qq', bestfit=True, estimate_ci=False, return_best_fit_results=True,problabel='Standard Normal Quantiles',
		datalabel=r'Parameter',
		scatter_kws={'markersize': 10, 'linestyle': 'none', 'label': None})
		slope, intercept, rcoef, prob, sterrest = stats.linregress(r['x'],r['y'])
		plt.text(1.5,max(r['y']), "$R^2_{"+str(method)+"}=%1.2f$" % rcoef**2)
		plt.legend(loc='lower right', fontsize='medium')
		plt.title('QQ-plot')
		if(save_plot != None):
			try:
				checkPath(save_plot)
				plt.savefig(save_plot+'qqplot.pdf', bbox_inches='tight', format='pdf', dpi=1200 )
			except NameError:
				print('No path to save the plots has been chosen')
		else:
			print('No path has been defined\nTrying to use GUI backend...')
			plt.show()
		plt.close()

		return

	def varplot(self,fds,var_source = None, backend = None, save_plot = None):
		"""
		Methods
		-------
		varplot(self,fds, save_plot = None):
			Plot all the IV curves for a common variability source.

		Parameters
		----------
		fds : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
		var_source : string or None, optional
			Information of the variability applied to device
		save_plot : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.

		"""
		import matplotlib
		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass
		
		fig, ax1 = plt.subplots()	
		ax2 = ax1.twinx()	
		for i in range(len(fds.dataset)):
			try:		
				ax1.plot(fds.dataset[i][:,0][:],fds.dataset[i][:,1][:],lw =0.4,color='red', label='log')
				ax2.plot(fds.dataset[i][:,0][:],fds.dataset[i][:,1][:],lw =0.4,color='green', label='Lin')
				ax1.set_yscale('log')
				ax1.set_ylabel(r'$I_{D}[A] $',labelpad=5, fontsize=35)
				ax1.set_xlabel(r'$V_{G}[V] $',labelpad=5, fontsize=35)
				ax1.tick_params('x', labelsize=24)
				ax1.set_xlim([0,1])
				ax1.set_ylim([1e-10,2e-4])
				ax1.tick_params('y', labelsize=24)
				ax1.tick_params('y', width=2)
				if(var_source != None):
					ax1.set_title(var_source, fontsize=35)
				ax2.yaxis.set_visible(False)
				# ax2.tick_params('y', labelsize=24)
				# ax2.tick_params('y', width=2)
				plt.tight_layout() 	
			except(TypeError, ValueError):
				pass					
		if(save_plot != None):
			try:
				checkPath(save_plot)
				plt.savefig(save_plot+'varplot_%s.pdf' %str(var_source), bbox_inches='tight', format='pdf', dpi=1200 )
			except NameError:
				print('No path to save the plots has been chosen')
		else:
			print('No path has been defined\nTrying to use GUI backend...')
			plt.show()
		plt.close()
		return

	def varplotlogCC(self, fds,var_source = None, cc_criteria = None, backend = None, save_plot = None):
		"""
		Methods  @Julian 2021/02
		-------
		varplotlogCC(self,fds, save_plot = None):
			Plot all the IV curves for a common variability source in logarithmic scale showing the Constant current criteria

		Parameters
		----------
		fds : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
		save_plot : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.
		var_source : string or None, optional
			Information of the variability applied to device.
		cc_criteria : float
			Current value used to extract threshold voltage with this criteria.


		"""
		import matplotlib
		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass

		plt.close()
		fig, ax = plt.subplots()	
		for i in range(len(fds.dataset)):
			try:		
				ax.plot(fds.dataset[i][:,0][:],fds.dataset[i][:,1][:],lw =0.4,color='red', label='log')
				ax.set_yscale('log')
				ax.set_ylabel(r'$I_{D}[A] $',labelpad=5, fontsize=35)
				ax.set_xlabel(r'$V_{G}[V] $',labelpad=5, fontsize=35)
				ax.tick_params('x', labelsize=24)
				ax.set_xlim([0,1])
				ax.set_ylim([1e-10,2e-4])
				ax.tick_params('y', labelsize=24)
				ax.tick_params('y', width=2)
				if(var_source != None):
					plt.annotate(var_source, xy=(0.45, 0.2), xycoords="axes fraction", fontsize=25) # Introduce label of variability source applied
				if(cc_criteria != None):
					matplotlib.pyplot.axhline(cc_criteria, linestyle='--') # Blue horizontal line set for Constant Current criteria with Icc
				plt.tight_layout() 	
			except(TypeError, ValueError):
				pass					
		if(save_plot is not None):
			try:
				checkPath(save_plot)
				plt.savefig(save_plot+'varplotlogCC_%s.pdf' %str(var_source), bbox_inches='tight', format='pdf', dpi=1200 )
			except NameError:
				print('No path to save the plots has been chosen')
		else:
			print('No path has been defined\nTrying to use GUI backend...')
			plt.show()
		
		return

	def calib(self,fds1,fds2,backend = None, save_plot = None):

		import matplotlib
		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass

		for i in range(len(fds1.dataset)):
			try:
				plt.close()
				fig, ax1 = plt.subplots()
				ax2 = ax1.twinx()
				ax1.plot(fds1.dataset[i][:,0][:],fds1.dataset[i][:,1][:],'o',linestyle= '-.',markerfacecolor="None",color='red', label=r'$\mathrm{V_{D}}=$'+str(fds1.drain_bias_value)+' V')
				ax2.plot(fds1.dataset[i][:,0][:],fds1.dataset[i][:,1][:],'o',linestyle= '-.',markerfacecolor="None",color='red')
				ax1.plot(fds2.dataset[i][:,0][:],fds2.dataset[i][:,1][:],'s',linestyle= '-.',markerfacecolor="None",color='blue', label=r'$\mathrm{V_{D}}=$'+str(fds2.drain_bias_value)+' V')
				ax2.plot(fds2.dataset[i][:,0][:],fds2.dataset[i][:,1][:],'s',linestyle= '-.',markerfacecolor="None",color='blue')
				ax1.set_yscale('log')
				ax1.set_ylabel('$I_{D} [\mu A/\mu m]$')
				ax1.set_xlabel('$V_{G} [V]$')
				ax1.set_title('Calibration plot')
				ax1.legend()
				plt.tight_layout()
			except(TypeError, ValueError):
				plt.close()	
		if(save_plot != None):
			try:
				checkPath(save_plot)
				plt.savefig(save_plot+'calibration.pdf', bbox_inches='tight', format='pdf', dpi=1200 )
			except NameError:
				print('No path to save the plots has been chosen')
		else:
			print('No path has been defined\nTrying to use GUI backend...')
			plt.show()
		plt.close()
		return

	def fomplot(self,i,fds1, fom = None,  currents = None, voltages = None, parameter = None,method = None, cc_criteria = None,parameter_vth = None, vg_ext = None, curve_high = None, curve_low = None, vth_high = None, vth_low = None, corriente_low = None,backend = None, save_plot = None, A=None, B=None, vg_start = None, vg_end = None, vt_sd_medio = None):
		"""
		Methods
		-------
		fomplot(i, fom = None,  voltages = None, currents = None, parameter = None,method = None, cc_criteria = None,parameter_vth = None,
		vg_ext = None, curve_high = None, curve_low = None, vth_high = None, vth_low = None, corriente_low = None, save_plot = None, A=None, B=None, vg_start = None, vg_end = None, vt_sd_medio = None):

			Plot the most common figures of merit of a semiconductor's IV curve.

		Parameters
		----------
		fom : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
		currents : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.
		voltages : path or None, optional
			Path indicating the folder where the user wishes to save the generated plots.			
		backend : str
			String containing the name of the backend chosen to either plot or save the plots. The backends available are:
			'Agg', which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tools for visualizing the plots.'TkAgg' requires the package python3-tk installed in order to run.
		"""
		import matplotlib

		if(backend == None) or (backend == 'Agg'):
			matplotlib.use('Agg')
		elif(backend == 'TkAgg'):
			matplotlib.use('TkAgg')
		else:
			pass

		if(currents != None):
			if(len(currents)<2):
				raise Exception('Not enough data for a variability curve')
		else:
			if(len(curve_high)<2) or (len(curve_low)<2):
				raise Exception('Not enough data for a variability curve')

		if(fom == None):
			raise Exception('The figure of merit to plot wasn\'t defined!')

		#Plot parameters
		curve_color = 'dodgerblue' #Color of the curve
		curve_label = 'Simulated Data' #Label of the curve
		x_label = r'$\mathrm{V_{G}} $' #Name of the x label
		y_label = r'$\mathrm{I_{D}} $' #Name of the y label
		label_pad = 20 #Padding applied to the x and y labels
		yscale = "log" #Type of scale["linear", "log", "symlog", "logit", ...]
		txt_fontsize = 15 #Fontsize of the plot
		txt_family = 'serif' #Font family of the plot
		txt_style = 'normal' #Font style of the plot

		if(fom == 'vth'):

			if(method == 'SD') or(method == None):

				title=r'$\mathrm{V_{TH}} $ SD extraction method ->Curve %s' % str(i) #Title of the plot
				SD_color = 'pink' #Color of the Second Derivative curve
				SD_linestyle = '--' #Line Style of the Second Derivative curve
				SD_label = 'SD ($d^{2}I_{D}/dV_{G}^{2}$)' #Label of the Second Derivative curve
				line_color = 'blue'
				line_style = '-.'
				line_label = r'$\mathrm{V_{TH}}$' #Label of the FoM vertical line	

				plt.close()
				fig, ax1 = plt.subplots()
				ax2 = ax1.twinx()
				curve = np.column_stack((voltages,currents))
				if(fds1.drain_bias_label=='High'):
					curve[:,1] = np.power( curve[:,1],0.5)
				ax1.plot(curve[:,0], curve[:,1],color = curve_color,label = curve_label)
				ax1.set_yscale(yscale)
				ax1.set_xlabel(x_label,labelpad=label_pad)
				ax1.set_ylabel(y_label,labelpad=label_pad)
				textstr = '$V_{T}=%.3f$'%(parameter)
				props = dict(boxstyle="square,pad=0.1", fc='w', alpha=1,ec='k', lw =2)
				ax1.text(0.05, 0.95, textstr, transform=ax1.transAxes, fontsize=txt_fontsize,  va='top', bbox=props,family = txt_family, style = txt_style)
				d2_curve = get_diff(curve, order = 2)
				d2_interp_x, d2_interp_y = interpol(d2_curve[:,0], d2_curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
				d2_interp = np.column_stack((d2_interp_x, d2_interp_y))
				ax2.plot(d2_interp[:, 0],d2_interp[:, 1], color=SD_color, linestyle= SD_linestyle ,label = SD_label)
				ax2.yaxis.set_visible(False) #The scale of the second derivative is removed
				ax1.axvline(x = parameter,color=line_color,label = line_label) #The value of the vth is represented with a vertical line
				ax1.legend(loc='best',fontsize = txt_fontsize)
				plt.title(title)
				plt.tight_layout()
				plt.show()
				if(save_plot != None):
					try:
						checkPath(save_plot)
						plt.savefig(save_plot+'vth_SD_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
					except NameError:
						print('No path to save the plots has been chosen')
				else:
					print('No path has been defined\nTrying to use GUI backend...')
					plt.show()
				plt.close()

			if(method == 'CC'):
				title=r'$\mathrm{V_{TH}} $ CC extraction method ->Curve %s' % str(i) #Title of the plot
				CC_color = 'pink' #Color of the CC curve
				CC_linestyle = '--' #Line Style of the CC curve
				CC_label = 'CC criteria' #Label of the CC curve
				line_color = 'blue'
				line_style = '-.'
				line_label = r'$\mathrm{V_{TH}}$' #Label of the FoM vertical line					

				plt.close()
				fig, ax1 = plt.subplots()			
				ax1.plot(voltages,currents,color = curve_color,label = curve_label)
				ax1.set_yscale(yscale)					
				ax1.set_xlabel(x_label,labelpad=label_pad)
				ax1.set_ylabel(y_label,labelpad=label_pad)
				textstr = '$V_{T}=%.3f$'%(parameter)
				props = dict(boxstyle="square,pad=0.1", fc='w', alpha=1,ec='k', lw =2)
				ax1.text(0.05, 0.95, textstr, transform=ax1.transAxes, fontsize=txt_fontsize,  va='top', bbox=props,family = txt_family, style = txt_style)
				ax1.axhline(y=cc_criteria, color=CC_color, linestyle=CC_linestyle, label = CC_label)
				ax1.legend(loc='best',fontsize =txt_fontsize)
				plt.title(title)
				plt.tight_layout()				
				if(save_plot != None):
					try:
						checkPath(save_plot)
						plt.savefig(save_plot+'vth_CC_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
					except NameError:
						print('No path to save the plots has been chosen')
				else:
					print('No path has been defined\nTrying to use GUI backend...')
					plt.show()
				plt.close()

			if(method == 'TD'):
				title=r'$\mathrm{V_{TH}} $ TD extraction method ->Curve %s' % str(i) #Title of the plot
				TD_color = 'pink' #Color of the TD curve
				TD_linestyle = '--' #Line Style of the TD curve
				TD_label = 'SD ($d^{3}I_{D}/dV_{G}^{3}$)' #Label of the TD curve
				line_color = 'blue'
				line_style = '-.'
				line_label = r'$\mathrm{V_{TH}}$' #Label of the FoM vertical line	
				path_to_save = str(save_plot) + 'vth_TD'+'%03d'%i+'.pdf' #Path to save to plot to

				plt.close()
				fig, ax1 = plt.subplots()
				ax2 = ax1.twinx()
				curve = np.column_stack((voltages,currents))
				if(fds1.drain_bias_label=='High'):
					curve[:,1] = np.power( curve[:,1],0.5)
				ax1.plot(curve[:,0], curve[:,1],color = curve_color,label = curve_label)
				ax1.set_yscale(yscale)					
				ax1.set_xlabel(x_label,labelpad=label_pad)
				ax1.set_ylabel(y_label,labelpad=label_pad)
				textstr = '$V_{T}=%.3f$'%(parameter)
				props = dict(boxstyle="square,pad=0.1", fc='w', alpha=1,ec='k', lw =2)
				ax2.text(0.05, 0.95, textstr, transform=ax1.transAxes, fontsize=14,  va='top', bbox=props,family = txt_family, style = txt_style)

				d3_curve = get_diff(curve, order = 3)
				d3_interp_x, d3_interp_y = interpol(d3_curve[:,0], d3_curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
				ax2.plot(d3_interp_x, d3_interp_y, color=TD_color, linestyle= TD_linestyle ,label = TD_label)
				ax2.yaxis.set_visible(False) #The scale of the second derivative is removed
				ax1.axvline(x = parameter,color=line_color, linestyle = line_style, label = line_label) #The value of the vth is represented with a vertical line
				ax1.legend(loc='best',fontsize = txt_fontsize)
				plt.title(title)
				plt.tight_layout()			
				if(save_plot != None):
					try:
						checkPath(save_plot)
						plt.savefig(save_plot+'vth_TD_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
					except NameError:
						print('No path to save the plots has been chosen')
				else:
					print('No path has been defined\nTrying to use GUI backend...')
					plt.show()
				plt.close()

			if(method == 'LE'):
				title=r'$\mathrm{V_{TH}} $ LE extraction method ->Curve %s' % str(i) #Title of the plot
				LE_color = 'pink' #Color of the LE curve
				LE_deriv_color = 'red' #Color of the LE curve
				LE_linestyle = '--' #Line Style of the LE curve
				LE_label = 'LE criteria' #Label of the LE curve
				LE_deriv_label ='1rst deriv' #Label of the first derivative
				line_color = 'blue'
				line_style = '-.'
				line_label = r'$\mathrm{V_{TH}}$' #Label of the FoM vertical line				

				plt.close()
				fig, ax1 = plt.subplots()
				ax2 = ax1.twinx()
				curve = np.column_stack((voltages, currents))
				if(fds1.drain_bias_label=='High'):
					curve[:,1] = np.power( curve[:,1],0.5)
				d1_curve = get_diff(curve, order = 1)
				d1_interp_x, d1_interp_y = interpol(d1_curve[:,0], d1_curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
				x_interp, y_interp = interpol(voltages, currents,strategy = fds1.interpolation, n = 1000,s = 0)				
				ax1.plot(voltages,currents,color = curve_color,label = curve_label)
				fit = A*x_interp+B
				vt_index_le_fit = find_closest(fit,0)
				ax1.plot(x_interp, A*x_interp+B,color=LE_color,linestyle =LE_linestyle, label=LE_label)
				ax2.plot(d1_interp_x, d1_interp_y,color=LE_deriv_color,linestyle =LE_linestyle, label=LE_deriv_label)
				ax1.set_xlabel(x_label,labelpad=label_pad)
				ax1.set_ylabel(y_label,labelpad=label_pad)
				textstr = '$V_{T}=%.3f$'%(parameter)
				props = dict(boxstyle="square,pad=0.1", fc='w', alpha=1,ec='k', lw =2)
				ax2.text(0.05, 0.95, textstr, transform=ax1.transAxes, fontsize=14,  va='top', bbox=props,family = txt_family, style = txt_style)
				ax1.legend(loc='best')
				ax2.yaxis.set_visible(False)
				ax1.axvline(x = parameter,color=line_color, linestyle = line_style, label = line_label) #The value of the vth is represented with a vertical line
				ax1.axhline(y = 0,color=line_color, linestyle = line_style, label = line_label) #The value of the vth is represented with a vertical line				
				plt.title(title)
				plt.tight_layout()
				if(save_plot != None):
					try:
						checkPath(save_plot)
						plt.savefig(save_plot+'vth_LE_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
					except NameError:
						print('No path to save the plots has been chosen')
				else:
					print('No path has been defined\nTrying to use GUI backend...')
					plt.show()
				plt.close()

		if(fom == 'ioff'):
			title=r'$\mathrm{I_{OFF}[A]}$ extraction ->Curve %s' % str(i) #Title of the plot
			IOFF_color = 'pink' #Color of the IOFF curve
			IOFF_linestyle = '--' #Line Style of the IOFF curve
			IOFF_label = r'$\mathrm{I_{OFF}}[A]$' #Label of the IOFF curve
			line_color = 'blue'
			line_style = '-.'
			line_label = r'$\mathrm{I_{OFF}}$' #Label of the FoM vertical line
				
			plt.close()
			fig, ax1 = plt.subplots()
			ax1.set_yscale(yscale)	
			ax1.plot(voltages,currents,color = curve_color,label = curve_label)
			ax1.set_xlabel(x_label,labelpad=label_pad)
			ax1.set_ylabel(y_label,labelpad=label_pad)			
			ax1.axhline(y=float(parameter), color=IOFF_color, linestyle=IOFF_linestyle, label = IOFF_label)
			index = find_closest(currents, parameter)
			ax1.axvline(x = float(voltages[index]), color=line_color, linestyle=line_style)
			if(vg_ext != None):
				ax1.axvline(x = vg_ext, color=line_color, linestyle=line_style, label = line_label)
			ax1.legend(loc='best',fontsize =25)
			plt.title(title)
			plt.tight_layout()

			if(save_plot != None):
				try:
					checkPath(save_plot)
					plt.savefig(save_plot+'IOFF_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
				except NameError:
					print('No path to save the plots has been chosen')
			else:
				print('No path has been defined\nTrying to use GUI backend...')
				plt.show()
			plt.close()

		if(fom == 'ion'):
			plt.close()
			fig, ax1 = plt.subplots()
			ax1.set_yscale("log")
			ax1.plot(voltages,currents,color='orangered',label ='Simulated Data')
			ax1.axhline(y=float(parameter), color='blue', linestyle='-.', label =r'$\mathrm{I_{ON}}[A]$')
			index = find_closest(currents, parameter)
			ax1.axvline(x = float(voltages[index]), color='blue', linestyle='-.')
			if(vg_ext != None):
				ax1.axvline(x = vg_ext, color='blue', linestyle='-.')
			ax1.set_ylabel(r'$\mathrm{I_{D}} $',labelpad=20)
			ax1.set_xlabel(r'$\mathrm{V_{G}} $',labelpad=20)
			ax1.legend(loc='best',fontsize =25)
			plt.title(r'$\mathrm{I_{ON}[A]}$ extraction ->Curve %s' % str(i))
			plt.tight_layout()

			if(save_plot != None):
				try:
					checkPath(save_plot)
					plt.savefig(save_plot+'ION_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
				except NameError:
					print('No path to save the plots has been chosen')
			else:
				print('No path has been defined\nTrying to use GUI backend...')
				plt.show()
			plt.close()

		if(fom == 'ss'):
			plt.close()
			fig, ax1 = plt.subplots()
			#ax1.set_yscale("log")
			plt.plot(voltages, np.log10(currents), 'o')
			center = find_closest(voltages,(voltages[0]+voltages[-1])/2)
			if(vg_start == None) and (type(vg_end) == float):
				index = find_closest(voltages,vg_end)
				slope, intercept, _, _, _ = stats.linregress(voltages[0:index],np.log10(currents[0:index]))
			elif(vg_start != None) and (vg_end != None) and (type(vg_end) == float) and (type(vg_start) == float):
				index_start = find_closest(voltages,vg_start)
				index_end = find_closest(voltages,vg_end)
				slope, intercept, _, _, _ = stats.linregress(voltages[index_start:index_end],np.log10(currents[index_start:index_end]))
			elif(vg_start != None) and (type(vg_start) == float):
				index_start = find_closest(voltages,vg_start)
				index = find_closest(voltages,vt_sd_medio)
				slope, intercept, _, _, _ = stats.linregress(voltages[index_start:index],np.log10(currents[index_start:index]))
			else:
				index = find_closest(voltages,vt_sd_medio)
				slope, intercept, _, _, _ = stats.linregress(voltages[0:index],np.log10(currents[0:index]))

			# print(1000/slope)
			dummy=np.linspace(0.0,voltages[center],1000)
			yss=dummy*slope+intercept
			plt.plot(dummy,yss,'-',color='r', label='SS')
			plt.ylabel(r'$\mathrm{I_{D}} $',labelpad=20)
			plt.xlabel(r'$\mathrm{V_{G}} $',labelpad=20)
			plt.title('Subthreshold Slope [mV/dec] ->Curve %s' % str(i))
			plt.tight_layout()

			if(save_plot != None):
				try:
					checkPath(save_plot)
					plt.savefig(save_plot+'SS_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
				except NameError:
					print('No path to save the plots has been chosen')
			else:
				print('No path has been defined\nTrying to use GUI backend...')
				plt.show()
			plt.close()

		if(fom == 'dibl'):
			vth_index_high = find_closest(curve_high[:,0],vth_high)
			corriente_high = curve_low[vth_index_high,1]
			vth_index_low = find_closest(curve_low[:,0],vth_low)
			corriente_low = curve_low[vth_index_low,1]

			plt.close()
			fig, ax1 = plt.subplots()
			ax1.set_yscale("log")
			ax1.plot(curve_low[:,0], curve_low[:,1], '-.', color='r')
			ax1.plot(curve_high[:,0], curve_high[:,1], '-.', color='b')
			ax1.axvline(x=vth_low, color='r', label ='Vt low')
			ax1.axhline(y=corriente_low, color='k')
			ax1.axvline(x=vth_high, color='b', label ='Vt high')
			ax1.set_xlabel(r'$\mathrm{V_{G}} $',labelpad=20)
			ax1.set_ylabel(r'$\mathrm{I_{D}} $',labelpad=20)
			plt.title('DIBL ->Curve %s' % str(i))
			ax1.legend(loc='lower right')
			plt.tight_layout()

			if(save_plot != None):
				try:
					checkPath(save_plot)
					plt.savefig(save_plot+'DIBL_{0}.pdf'.format(str(i)), bbox_inches='tight', format='pdf', dpi=1200 )
				except NameError:
					print('No path to save the plots has been chosen')
			else:
				print('No path has been defined\nTrying to use GUI backend...')
				plt.show()
			plt.close()

#----------------------------------------------------------------------------------------------------------------	

def interpol(x = None ,y = None, n = None, strategy = None, d = None, s = None):
	"""
	Wrapper function for interpolating imported data from a semiconductor's IV curve.

	Parameters
	----------
	x : array_like, shape (n,)
		1-d array containing values of the independent variable.
	y : array_like
		Array containing values of the dependent variable.
		It can have arbitrary number of dimensions, but the length along axis
		must match the length of x. Values must be finite.
	strategy : str
		Keyword for defining the selected interpolation method: The list of available methods includes:
		'akima', 'pchip' and 'linear'.
	d : int
		Degree of the smoothing spline. Must be <= 5.
		Default is k=3, a cubic spline.
	s : float
		Positive smoothing factor used to choose the number of knots. 
	
	"""

	if(strategy == None) or (strategy == 'cubic_spline'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.spline_interpol(x, y, n, d, s)

	elif(strategy == 'akima'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.akima_interpol(x, y, n)

	elif(strategy == 'pchip'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.pchip_interpol(x, y, n)

	elif(strategy == 'linear'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.lin_interpol(x, y, n)

	return x_interp, y_interp




# -*- coding: utf-8 -*-
"""
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
fds.py 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module includes the routines used to import data into a FoMpy Dataset. Some useful
examples showing how to import the data using these functions can be seen below:

Example
-------

If the file parser is defined a simple voltage-current file is defined as input::

	import fompy
	path_file = './data/default/'
	fds = fompy.dataset(path, parser=fompy.file)

or if an array is passed as input::

	import fompy
	#Here several arrays are defined
	arr1 =np.array([[0.00e+00, 1.00e-09],[1.00e-01, 2.20e-08],[2.00e-01, 3.20e-07],[3.00e-01, 2.74e-06],[4.00e-01, 9.90e-06],[5.00e-01, 2.20e-05],[6.00e-01, 3.22e-05],[7.00e-01, 4.16e-05],[8.00e-01, 5.23e-05],[9.00e-01, 6.04e-05],[1.00e+00, 6.60e-05]])
	arr2 =np.array([[0.00e+00, 1.00e-09],[1.00e-01, 2.15e-08],[2.00e-01, 3.18e-07],[3.00e-01, 2.72e-06],[4.00e-01, 9.85e-06],[5.00e-01, 2.12e-05],[6.00e-01, 3.16e-05],[7.00e-01, 4.10e-05],[8.00e-01, 5.46e-05],[9.00e-01, 6.15e-05],[1.00e+00, 6.57e-05]])
	arrays = np.stack((arr1, arr2)) #Here the arrays are put together
	fds = fompy.dataset(arr = arrays, parser=fompy.array)

	#Also for a single IV curve
	x = ([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
	y = ([1.00e-09, 2.20e-08, 3.20e-07, 2.74e-06, 9.90e-06, 2.20e-05, 3.22e-05, 4.16e-05, 5.23e-05, 6.04e-05, 6.60e-05])
	fds = fompy.iv(arr = (x,y), parser=fompy.curve)



Moreover, when parsers for another file-types are be selected::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)

	path_file_mc = './data/mc_data/'
	fds = fompy.dataset(path_file_mc, parser=fompy.MC)


If the user wishes to remove several IV curves included in the parent folder, two options called
'interval' and 'exclude' can be passed, so the indexes defined are removed from the FoMpy Dataset::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB, exclude=[5,6])
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB, interval=[0,8])

"""

from abc import ABCMeta, abstractmethod
import os, glob
from enum import Enum
import numpy as np
from scipy.stats import mode

#--------------------------------------------------------------------------------------------------------------
class _FompyDatasetStrategy(metaclass=ABCMeta):
	"""
	Abstract class containing the most important attributes of a semiconductor's IV curve.

	"""
	def __init__(self, **kwargs):
		self.dataset = []
		self.n_sims = 0
		self.sanity_array = []
		self.norm = 1
		self.ext_method = 'SD'
		self.drain_bias_label = None
		self.drain_bias_value = None
		self.interpolation = 'cubic_spline'
		self.filter = None
		self.path_subdirs = []


	def print_parameters(self):
		"""
		Prints the details of a FompyDataset.
		"""		
		print("\nNumber of simulations: ", self.n_sims)
		print("Normalization value: ", self.norm)
		print("Extraction method: ", self.ext_method)
		print("Drain Bias label: ", self.drain_bias_label)
		print("Drain Bias value: ", self.drain_bias_value)
		print("Interpolation: ", self.interpolation)
		print("Filtering: ", self.filter, "\n")



#--------------------------------------------------------------------------------------------------------------
class FompyDataset(_FompyDatasetStrategy):
	"""
	Class containing the simulated IV curves and their parameters.

	...

	Attributes
	----------
	dataset : array: double[]
		List containing all IV curves
	n_sims : int
		Number of simulated IV curves
	sanity_array : array: double[]
		Array of ones by default. If a simulation has failed, its index is converted to zero.
	norm : double
		Normalization value applied to the IV curve
	ext_method : str
		FompyDataset default method used to extrad the figures of merit
	drain_bias_label : str
		Either high or low.
	drain_bias_value : double
		Drain voltage value used to simulate the IV curves.
	interpolation : str
		IV curve interpolation method. Can either be 'cubic_spline',
		'akima','pchip' or 'linear'.
	filter : str
		IV curve filter method. So far only the polar filter has been implemented ('polar_filter').

	Methods
	-------
	print_parameters
		Prints all the non-array attributes.

	"""
	def __init__(self,**kwargs):
		_FompyDatasetStrategy.__init__(self, **kwargs)
		for key in kwargs:
			setattr(self, key, kwargs[key])

#--------------------------------------------------------------------------------------------------------------
class dataDAO(metaclass=ABCMeta):
	"""
	Data Acces Object Interface used to import the simulated IV curves to a FompyDataset.
	"""	

	@abstractmethod
	def load(self):
		pass
	@abstractmethod
	def save(self):
		pass

class daoFile(dataDAO, FompyDataset):
	"""
	Data Acces Object used to import from a set of files the simulated IV curves to a FompyDataset.

	...

	Attributes
	----------
	parser : str
		Type of parser, user-defined, that imports the simulated IV curves in a specific format.

	"""	

	def __init__(self, parser = None):
		self.parser = parser

	def load(self, path = None,arr = None,iv = None, parser = None, interval = None, exclude = None):
		"""
		Methods
		-------
		load(path,filename_user = None, parser = None, interval = None, exclude = None)
			Class method that extracts :math:`V_{TH}` of a semiconductor's IV curve.

		Parameters
		----------
		path : array_like, shape (n,)
			1-d array containing values of the independent variable.
		arr : array_like
			Array of data containing one or more semiconductor's IV curves.
		iv : array_like
			Single IV curve.
		parser : function
			Function that implements how the data is imported to a Fompy Dataset. The list of available functions includes:
			'file','array','JCJB' and 'MC'.
		interval : array_like
			List of two int values: start(index of the first simulation to load into the Fompy Dataset)
			and end(index of the last simulation to load into the FompyDataset)
		exclude : array_like
			Index values of simulations to exclude.
		skiprows : int
			Number of rows to skip at the begining of a file. 0 rows are skipped by default.
		comments : str
			All the lines starting with this character are considered comments.
			'#' is used by default.

		"""
		fds = FompyDataset()
		fds.drain_bias_label = 'High' 
		fds.drain_bias_value = 0.7
		if(path is not None):
			# path_subdirs = []
			path_filenames = []
			for dirname, dirnames, filenames in os.walk(path):
				# print path to all subdirectories.
				for subdirname in dirnames:
					fds.path_subdirs.append(os.path.join(dirname, subdirname))
				# print path to all filenames.
				for filename in filenames:
					path_filenames.append(os.path.join(dirname, filename))

			fds.path_subdirs = sorted(fds.path_subdirs)
			path_filenames = sorted(path_filenames)
			# print(path_filenames)
			# print(path_subdirs)


			#MC version is not for variability studies (several folders for IV curve)
			if(parser is not MC):
				for i in range(len(fds.path_subdirs)):
					try:
						if(fds.path_subdirs[i] in str(path_filenames)):
							fds.sanity_array.append(1)
						else:
							fds.sanity_array.append(-1)
					except IndexError:
						fds.sanity_array.append(-1)
			else:
				fds.sanity_array.append(1)
	
		if(parser is file):
			parser(fds, path, fds.path_subdirs, path_filenames, interval, exclude)
		elif(parser is JCJB):
			parser(fds, path, fds.path_subdirs, path_filenames, interval, exclude)
		elif(parser is array):
			parser(fds, arr)
		elif(parser is curve):
			parser(fds, iv)
		elif(parser is MC):
			parser(fds, MC, fds.path_subdirs, path_filenames, interval, exclude)
		elif(parser is Silvaco):
			parser(fds, path)			
		else:		
			print('No parser has been defined')

		if(path is not None):
			fds = exclude_indexes(fds,interval, exclude)
		return fds

	def save(self, fds, path):
		length = len(fds.dataset)
		f = open(str(path), "w+")
		for i in range(length):
			if(fds.dataset[i] is not np.nan):
				f.write("#Simulation id {0}, Vd = [{1}]\n".format(i+1, fds.drain_bias_value))
				f.write("#Vg[V]\tId[V]\n")
				arr = np.array2string(fds.dataset[i][:], separator="\t")
				f.write(arr+"\n")
			else:
				f.write("Simulation id {0}\n".format(i+1))
				f.write("FAILURE TO LOAD THE DATA\n")
		f.close()


#--------------------------------------------------------------------------------------------------------------
def file(fds, path, path_subdirs, path_filenames, interval, exclude):
	"""Function that imports the simulated data from a given file and
	stores it into a FoMpy Dataset. The format of this file is asummed to have comments
	starting with '#', without a header and two columns, one for the voltage and one for the currents, separated
	by '\t' delimiters.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	path : str
		Parent path where the simulations are stored
	path_subdirs : str
		List of alphabetically sorted subdirectories found inside the parent directory.
	path_filenames : str
		List of alphabetically sorted files found inside the parent directory.
	interval : array_like
		List of two int values: start(index of the first simulation to load into the Fompy Dataset)
		and end(index of the last simulation to load into the FompyDataset)
	exclude : array_like
		Index values of simulations to exclude.
		
	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the FoMpy Dataset.

	"""
	extension = [".txt", ".dat", ".out", ".csv"]
	
	print(path_filenames)
	for item in path_filenames:
		# print(item)
		file_name = os.path.basename(item)
		for ext in extension:
			if(file_name.endswith(ext) is True):
				print('Reading data from %s file -> %s' % (ext ,file_name))
				data_temp = np.loadtxt(item, skiprows=1, unpack=True, delimiter='\t', comments='#')
				a, b = data_temp[0], data_temp[1]
				arr = np.column_stack((a, b))
				fds.dataset.append(arr)

	fds.n_sims = len(fds.dataset)

	return fds

#--------------------------------------------------------------------------------------------------------------
def array(fds, arr= None):
	"""Function that imports the simulated data from a given an array and
	stores it into a FoMpy Dataset.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	arr : array_like
		Array of data containing one or more semiconductor's IV curves.
	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the FoMpy Dataset.

	"""
	if(arr is not None):
		for i in range(len(arr)):
			fds.dataset.append(arr[i])
		fds.n_sims = len(fds.dataset)
	else:
		print('No dataset has been defined!')

	return fds

def curve(fds, iv = None):
	"""Function that imports the simulated data from a single array and
	stores it into a FoMpy Dataset.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	iv : array_like
		Single IV curve.
	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the FoMpy Dataset.

	"""

	#ANT.2020-10-05 Previous version, create a wrong array[(0.0, 1e-09), 
	# if(iv is not None):
	# 	for i in range(len(iv[0])):
	# 		fds.dataset.append((iv[0][i], iv[1][i]))
	# 	fds.n_sims = 1
	# else:
	# 	print('No dataset has been defined!')

	if(iv is not None):
		dummy_list = []
		for i in range(len(iv[0])):
			dummy_list.append((iv[0][i], iv[1][i]))
		fds.dataset.append(np.array(dummy_list))
		fds.n_sims = 1
	else:
		print('No dataset has been defined!')

	return fds

#--------------------------------------------------------------------------------------------------------------
def JCJB(fds, path, path_subdirs, path_filenames, interval, exclude):
	"""Function that imports the simulated data from a JCJB file and
	stores it into a FoMpy Dataset.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	path : str
		Parent path where the simulations are stored
	path_subdirs : str
		List of alphabetically sorted subdirectories found inside the parent directory.
	path_filenames : str
		List of alphabetically sorted files found inside the parent directory.
	interval : array_like
		List of two int values: start(index of the first simulation to load into the Fompy Dataset)
		and end(index of the last simulation to load into the FompyDataset)
	exclude : array_like
		Index values of simulations to exclude.

	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the FoMpy Dataset.

	"""

	try:
		for path in path_filenames:
			if('JCJB.dat.0' in path):
				data_temp = np.loadtxt(path)
				a, b, c = data_temp[:,0], data_temp[:,1], data_temp[:,2]
				condition = np.mod(a, np.amax(a))==0
				vd_temp = np.copy(np.extract(condition, a))
				vg_temp = np.copy(np.extract(condition, b))
				id_temp = np.copy(np.extract(condition, c))
				Vd = np.copy(vd_temp[np.argmin(vg_temp):len(vg_temp)])
				Vg = np.copy(vg_temp[np.argmin(vg_temp):len(vg_temp)])
				Id = np.copy(id_temp[np.argmin(vg_temp):len(vg_temp)])
				arr = np.column_stack((Vd, Vg, Id))
				arr_f = np.column_stack((arr[:,1], arr[:,2]))
				vd = np.unique(Vd)
				fds.dataset.append(arr_f)
				fds.drain_bias_value = vd
	except (ValueError, IndexError):
		print('An error has ocurred during the JCJB data load')

	return fds
#--------------------------------------------------------------------------------------------------------------
def MC(fds, path, path_subdirs, path_filenames, interval, exclude):
	"""Function that imports the simulated data from a MC set of files and
	stores it into a FoMpy Dataset.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	path : str
		Parent path where the simulations are stored
	path_subdirs : str
		List of alphabetically sorted subdirectories found inside the parent directory.
	path_filenames : str
		List of alphabetically sorted files found inside the parent directory.
	interval : array_like
		List of two int values: start(index of the first simulation to load into the Fompy Dataset)
		and end(index of the last simulation to load into the FompyDataset)
	exclude : array_like
		Index values of simulations to exclude.
	skiprows : int
		Number of rows to skip at the begining of a file. 0 rows are skipped by default.
	comments : str
		All the lines starting with this character are considered comments.
		'#' is used by default.

	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the FoMpy Dataset.

	"""

	index=0
	for i in fds.path_subdirs:
		try:
			if('Graphs' in i):
				fds.path_subdirs.pop(index)
		except IndexError:
			pass
		index = index + 1

	i = 0
	time_arr = []
	avg_curr_arr = []
	vg_arr = []
	if (interval is not None):
		ran = range(interval[0], interval[1])
	for path in path_filenames:
		if('fichero_particula' in path):
			try:
				time, avg_curr = np.genfromtxt(path, unpack=True, usecols = (0,11))
				time_arr.append(time)
				avg_curr_arr.append(np.abs(avg_curr))
			except ValueError:
				print('Non-valid data found in the file.')				
			i = i+1
		elif('voltajes' in path):
			vg, vd = np.genfromtxt(path, delimiter='\t',  skip_header = 1, usecols = (1,2), unpack=True)
			fds._drain_bias_value = vd
			vg_arr.append(vg)

	time_f = []
	avg_curr_f = []
	index_time = []
	for i in range(len(avg_curr_arr)):
		try:
			if(i==0):
				index_1 = np.argwhere(np.in1d(time_arr[i],np.intersect1d(time_arr[i], time_arr[i+1])) == True)
			index_2 = np.argwhere(np.in1d(time_arr[i],np.intersect1d(time_arr[i], time_arr[i+1])) == True)
			if(index_1[-1] <= index_2[-1]):
				index_time.append(index_1[-1])
			else:
				index_time.append(index_2[-1])
			min = np.min(index_time[-1])
		except IndexError:
			pass
	for i in range(len(avg_curr_arr)):
		try:
				time_f.append(time_arr[i][min])
				avg_curr_f.append(avg_curr_arr[i][min])										
		except IndexError:
			pass

	arr_f = np.column_stack((vg_arr, np.abs(avg_curr_f)))
	fds.dataset.append(arr_f)

	return fds
#--------------------------------------------------------------------------------------------------------------
def Silvaco(fds, path_to_file):
	"""Function that imports the simulated data from a given Silvaco file and
	stores it into a FoMpy Dataset.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	path : str
		Parent path where the simulations are stored
		
	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the FoMpy Dataset.

	"""
	import csv
	print('Reading data from file -> %s' % (path_to_file))
	with open(path_to_file, 'r') as f:
    		temp = list(csv.reader(f, delimiter=","))

	import numpy as np
	
	vg = [float(temp[i+1][0]) for i in range(len(temp)-1)]
	i_drain = [float(temp[i+1][1]) for i in range(len(temp)-1)]

	arr = np.column_stack((vg, i_drain))
	fds.dataset.append(arr)
	fds.n_sims = len(fds.dataset)

	return fds
#--------------------------------------------------------------------------------------------------------------
def exclude_indexes(fds,interval, exclude):
	"""Function that generates an array of zeros and ones. If the simulation
	number 5 has failed, either because the folder is empty, or not enough voltages
	have been simulated, then fds.sanity_array[4] is set to zero.

	Parameters
	----------
	fds : FoMpy Dataset
		Structure of data containing the most important parameters of a semiconductor's IV curve.
	interval : array_like
		List of two int values: start(index of the first simulation to load into the Fompy Dataset)
		and end(index of the last simulation to load into the FompyDataset)
	exclude : array_like
		Index values of simulations to exclude.

	Returns
	-------
	fds : FoMpy Dataset
		Structure of data containing the sanity array.

	"""
	try:
		line_length = []
		index = 0
		for i in fds.sanity_array:
			try:
				if(i!=-1):
					line_length.append(len(fds.dataset[index][:,1][:]))
					index = index + 1
				else:
					line_length.append(0)
			except IndexError:
				line_length.append(0)
		index_crashed = line_length.index(0)
		fds.dataset.insert(index_crashed, np.nan)

		mode_length = mode(line_length)[0]
		for i in range(len(line_length)):
			if(line_length[i] != mode_length):
				fds.sanity_array.pop(i)
				fds.sanity_array.insert(i,-1)
	except ValueError:
		pass

	if (interval is not None):
		temp_dataset= fds.dataset[interval[0]:interval[1]]
		fds.dataset=[]
		fds.dataset=temp_dataset

	if (exclude is not None):
		if(len(exclude)!=0):
			index = 0
			for i in exclude:
				fds.dataset.pop(i-1)
				index = index+1
		else:
			print(exclude)
			fds.dataset.pop(exclude)

	fds.n_sims =len(fds.dataset)

	return fds




# -*- coding: utf-8 -*-
"""
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
fom.py 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module includes the routines used to extract several commonly used FoMs
in semiconductor simulations.

Example
-------

After the data is imported into a FoMpy Dataset, the user may use several routines implemented in FoMpy to extract the 'vth', 'ioff', 'ion', 'ss', 'ratio', 'power' or 'dibl'. Code examples explaining how to use them can be seen below.

In order to extrac the vth of a given IV curve, the following commands have to be used either in a script or 
in a python3 command line::

	import fompy
	path_file_high = './data/sim_FinFET_vd_high/'
	fds = fompy.dataset(path_file_JCJB, parser=fompy.JCJB)
	fds.drain_bias_label = 'High'
	vth_array = fompy.extract(fds, fom = 'vth')

If you are not using a JCJB file, the attribute of drain_bias_label has to be defined when using the data of a FoMpy dataset at high drain bias, otherwise low drain bias formulas will be used.
Also if a different FoM is needed, the user has to change the keyword fom, from 'vth' to another one contained in the list shown before.

Additionally, if the user wants to obtain the DIBL, two different FoMpy Datasets have to be imported::

	import fompy
	path_file_JCJB = './data/sim_FinFET_vd_high/'
	path_file_low = './data/sim_FinFET_vd_low/'

	fds_hdb = fompy.dataset(path_file_JCJB, parser='JCJB')
	fds_ldb = fompy.dataset(path_file_low, parser='JCJB')
	fds_hdb.drain_bias_value = 0.7
	fds_ldb.drain_bias_value = 0.05
	dibl_array = fompy.extract(fds_hdb, fds_ldb, fom = 'dibl')

"""

from abc import ABCMeta, abstractmethod
from enum import Enum
import numpy as np
from fompy.aux import find_closest, get_diff
from fompy.conditioning import interpolator
from fompy.plots import plotter
import os
from scipy.optimize import curve_fit

class _extractor(metaclass=ABCMeta):
	"""
	Abstract class containing the most common methods applied to an IV curve to extrat the figure of merit.

	"""
	@abstractmethod
	def extraction(self):
		pass
	@abstractmethod
	def save_results_to_file(self):
		pass
	@abstractmethod
	def plot(self):
		pass

#----------------------------------------------------------------------------------------------------------------	

class vth_ext(_extractor):
	"""
	Child class of _extractor that obtains the :math:`V_{TH}` figure of merit from a semiconductor's IV curve.

	"""
#----------------------------------------------------------------------------------------------------------------	
	def extraction(self,fds1,method=None, cc_criteria = None):
		"""
		Methods
		-------
		extraction(fds1,method=None, cc_criteria = None)
			Class method that extracts :math:`V_{TH}` of a semiconductor's IV curve.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for the extraction of any FoM.
		method : str
			Keyword indicating the desired method of extraction of the FoMs.The list of available methods includes:
			'SD', 'CC', 'TD' and 'LE'. If method is not defined the 'SD' is selected by default.
		cc_criteria : float, optional
			Float value used for the extraction of several FoMs using the constant current method.
		"""
		if(method == 'SD') or(method == None):
			length = len(fds1.dataset)
			parameter = []
			curves = []
			for i in range(length):
				try:
					curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))
					if(fds1.drain_bias_label=='High'):
						curve[:,1] = np.power( curve[:,1],0.5)
					d2_curve = get_diff(curve, order = 2)
					d2_interp_x, d2_interp_y = interpol(d2_curve[:,0], d2_curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
					d2_interp = np.column_stack((d2_interp_x, d2_interp_y))

					lower_bound=find_closest(d2_interp[:,0],d2_interp[50,0])
					higher_bound=find_closest(d2_interp[:,0],d2_interp[-100,0])
					vt_temp = np.argmax(d2_interp[lower_bound:higher_bound,1])
					if (d2_interp[vt_temp+lower_bound, 0] <= d2_interp[lower_bound, 0]):
						print('Vt value outside of the confidence interval for simulation', i)
					if (d2_interp[vt_temp+lower_bound, 0] >= d2_interp[higher_bound, 0]):
						print('Vt value outside of the confidence interval for simulation', i)
					vt_SD = d2_interp[vt_temp+lower_bound,0]

					try:
						vt_temp = np.argmax(d2_interp[lower_bound:higher_bound+lower_bound,1])
						vt_SD = d2_interp[vt_temp+lower_bound,0]
					except ValueError:
						print('Multiple indexes in curve {}'.format(i))

					if(fds1.drain_bias_label=='High'):
						curve[:,1] = np.power( curve[:,1],2)
					parameter.append(float(("%0.3f"%vt_SD)))
					curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))		
					curves.append(curve)

				except (TypeError, ValueError):
					parameter.append(np.nan)		
					curves.append(np.nan)

			return parameter, curves

		elif(method == 'CC'):
			length = len(fds1.dataset)
			parameter = []
			curves = []
			for i in range(length):
				try:		
					x_interp,y_interp = interpol(fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:],strategy = fds1.interpolation, n = 1000,s = 0)
					curve = np.column_stack((x_interp, y_interp))
					try:
						vt_index_cc=find_closest(curve[:,1],float(cc_criteria))
						vt_CC_float =curve[vt_index_cc,0]
						parameter.append(float(("%0.3f"%vt_CC_float)))
						curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))		
						curves.append(curve)
					except (TypeError):
						print('No CC criteria defined')
						break
				except (ValueError, TypeError):
					parameter.append(np.nan)		
					curves.append(np.nan)

			return parameter, curves

		if(method == 'TD') or(method == None):
			length = len(fds1.dataset)
			parameter = []
			curves = []
			for i in range(length):
				try:
					curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))
					if(fds1.drain_bias_label=='High'):
						curve[:,1] = np.power( curve[:,1],0.5)
					d3_curve = get_diff(curve, order = 3)
					d3_interp_x, d3_interp_y = interpol(d3_curve[:,0], d3_curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
					d3_interp = np.column_stack((d3_interp_x, d3_interp_y))

					lower_bound=find_closest(d3_interp[:,0],d3_interp[50,0])
					higher_bound=find_closest(d3_interp[:,0],d3_interp[-100,0])
					vt_temp = np.argmax(d3_interp[lower_bound:higher_bound,1])

					if (d3_interp[vt_temp+lower_bound, 0] <= d3_interp[lower_bound, 0]):
						print('Vt value outside of the confidence interval for simulation', i)
					if (d3_interp[vt_temp+lower_bound, 0] >= d3_interp[higher_bound, 0]):
						print('Vt value outside of the confidence interval for simulation', i)
					vt_TD = d3_interp[vt_temp+lower_bound,0]

					if(fds1.drain_bias_label=='High'):
						curve[:,1] = np.power( curve[:,1],2)						
					parameter.append(float(("%0.3f"%vt_TD)))
					curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))		
					curves.append(curve)

				except (TypeError, ValueError):
					parameter.append(np.nan)		
					curves.append(np.nan)

			return parameter, curves

		if(method == 'LE') or(method == None):

			def line(x, A, B):
				return A*x + B

			length = len(fds1.dataset)
			parameter = []
			curves = []
			A = []
			B = []
			for i in range(length):
				try:
					x_interp,y_interp = interpol(fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:],strategy = fds1.interpolation, n = 1000,s = 0)
					curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))
					
					if(fds1.drain_bias_label=='High'):
						curve[:,1] = np.power( curve[:,1],0.5)

					d1_curve = get_diff(curve, order = 1)
					d1_interp_x, d1_interp_y = interpol(d1_curve[:,0], d1_curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
					d1 = np.column_stack((d1_interp_x, d1_interp_y))

					index_lower_bound=find_closest(d1[:,0],d1[50,0])
					index_higher_bound=find_closest(d1[:,0],d1[-100,0])

					x_interp, y_interp = interpol(fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:],strategy = fds1.interpolation, n = 1000,s = 0)
					curve = np.column_stack((x_interp, y_interp))

					try:
						d1_init = find_closest(x_interp,d1[0,0])
						d1_max_index = np.argmax(d1[index_lower_bound:index_higher_bound,1])+index_lower_bound+d1_init
						vt_temp_d1 = find_closest(x_interp,d1[d1_max_index,0])
					except ValueError:
						print('Multiple indexes')

					corriente_vt_temp =curve[vt_temp_d1,1]
					lim1 = int(vt_temp_d1*0.9)
					lim2 = int(vt_temp_d1*1.1)

					A_i,B_i = curve_fit(line, curve[lim1:lim2,0],curve[lim1:lim2,1])[0]
					fit = A_i*x_interp+B_i
					vt_index_le_fit = find_closest(fit,0)

					if(fds1.drain_bias_label=='High'):
						try:
							vt_LE = x_interp[vt_index_le_fit]
							curve[:,1] = np.power( curve[:,1],2)
						except NameError:
							print('The drain bias value has not been defined')
					else:
						vd_low = fds1.drain_bias_value
						vt_LE = curve[vt_index_le_fit,0]+vd_low/2
				
					parameter.append(float(("%0.3f"%vt_LE)))
					
					curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))
					curves.append(curve)
					A.append(A_i)
					B.append(B_i)
				except (TypeError, ValueError, IndexError):
					parameter.append(np.nan)		
					curves.append(np.nan)
					A.append(np.nan)
					B.append(np.nan)					
			return parameter, curves, A, B

#----------------------------------------------------------------------------------------------------------------	
	def save_results_to_file(self,path, parameter,fds):
		"""
		save_results_to_file(path, parameter)
			Class method that saves the extracted :math:`V_{TH}` values.

		Parameters
		----------
		path : str
			Defines the path where the extracted results are saved to a file.
		parameter : array_like
			Array of extracted FoM values to be saved into the file.

		"""

		length = len(parameter)
		f = open(str(path), "w")
		f.write("#Sim_id\t#Vth[V]\n")
		for i in range(len(parameter)):
			if (str(type(parameter)) is not '<class \'numpy.ndarray\'>'):
				parameter = np.array(parameter)
			arr = np.array2string(parameter[i], separator="\t")
			directorio = fds.path_subdirs[i].split(sep='/')
			subdir = directorio[5]
			f.write("{0}\t{1}\n".format(subdir, arr))			
		f.close()
#----------------------------------------------------------------------------------------------------------------		
	def plot(self, fds1, parameter = None, method = None,cc_criteria = None, curves = None,backend= None, save_plot = None, A=None, B=None):
		"""
		plot(fds1, parameter = None, method = None,cc_crit = None, curves = None, save = None, A=None, B=None)
			Class method that plots the extracted :math:`V_{TH}` values.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for generating the plot of any FoM.
		parameter : array_like
			Array of extracted FoM values to be plotted.
		method : str
			Keyword indicating the desired method of extraction of the FoMs. The list of available methods includes: 'SD', 'CC', 'TD' and 'LE'. If method is not defined the 'SD' is selected by default.
		cc_criteria : float
				Current criteria used to extract vth with the CC criteria for the fomplot.
		curves : array_like
			Array of data containing the IV curves.
		backend : str
			String containing the name of the backend chosen to either plot or save the plots. The backends available are:
			'Agg', which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tools for visualizing the plots.
			'TkAgg' requires the package python3-tk installed in order to run.			
		save_plot : bool
			If True the generated plot is save to the defined path.
		A, B: float
			Parameters obtained during the vth LE extraction method used for the plots.

		"""

		temp_plot = plotter()
		for i in range(len(fds1.dataset)):
			try:
				if (method is not 'LE'):
					temp_plot.fomplot(i,fds1, fom = 'vth', voltages = curves[i][:,0], currents = curves[i][:,1],
				 	parameter = parameter[i], method = method,cc_criteria = cc_criteria,backend= backend, save_plot = save_plot)
				else:
					temp_plot.fomplot(i,fds1, fom = 'vth', voltages = curves[i][:,0], currents = curves[i][:,1],
				 	parameter = parameter[i], method = method,cc_criteria = cc_criteria,backend= backend, save_plot = save_plot, A = A[i], B=B[i])
			except (TypeError, ValueError):
				i = i+i


#----------------------------------------------------------------------------------------------------------------	

class ioff_ext(_extractor):
	"""
	Child class of _extractor that obtains the :math:`I_{OFF}` figure of merit from a semiconductor's IV curve.

	"""
#----------------------------------------------------------------------------------------------------------------	
	def extraction(self,fds1 = None,vg_ext = None):
		"""
		Methods
		-------
		extraction(fds1,method=None, cc_criteria = None)
			Class method that extracts :math:`I_{OFF}` of a semiconductor's IV curve.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for the extraction of any FoM.
		vg_ext : float
			Gate voltage value used to calculate IOFF at.
		"""
		parameter = []
		curves = []
		for i in range(len(fds1.dataset)):
			try:
				x_interp,y_interp = interpol(fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:],strategy=fds1.interpolation, n = 1000,s = 0)
				single_curve = np.column_stack((x_interp, y_interp))
				if(vg_ext is not 0) and (vg_ext is not None) :
					voltage_index = find_closest(x_interp, vg_ext)
					parameter.append(y_interp[voltage_index])
					curves.append(single_curve)

				else:
					parameter.append(y_interp[0])
					curves.append(single_curve)
			except (TypeError, ValueError):
				parameter.append(np.nan)		
				curves.append(np.nan)
		return parameter, curves
#----------------------------------------------------------------------------------------------------------------	
	def save_results_to_file(self,path, parameter,fds):
		"""
		save_results_to_file(path, parameter)
			Class method that saves the extracted :math:`I_{OFF}` values.

		Parameters
		----------
		path : str
			Defines the path where the extracted results are saved to a file.
		parameter : array_like
			Array of extracted FoM values to be saved into the file.

		"""
		length = len(parameter)
		f = open(str(path), "w+")
		f.write("#Sim_id\t#Ioff[A]\n")
		# f.write("##############################\n###")
		for i in range(len(parameter)):
			if (str(type(parameter)) is not '<class \'numpy.ndarray\'>'):
				parameter = np.array(parameter)
			arr = np.array2string(parameter[i], separator="\t")
			directorio = fds.path_subdirs[i].split(sep='/')
			subdir = directorio[5]
			f.write("{0}\t{1}\n".format(subdir, arr))			
		f.close()
#----------------------------------------------------------------------------------------------------------------	
	def plot(self, fds1, parameter = None,vg_ext = None, curves = None,backend= None, save_plot = None):
		"""
		plot(fds1, parameter = None, method = None,cc_crit = None, curves = None, save = None, A=None, B=None)
			Class method that plots the extracted :math:`I_{OFF}` values.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for generating the plot of any FoM.
		parameter : array_like
			Array of extracted FoM values to be plotted.
		vg_ext : float
			Gate voltage value used to calculate IOFF at.
		curves : array_like
			Array of data containing the IV curves.
		backend : str
			String containing the name of the backend chosen to either plot or save the plots. The backends available are:
			'Agg', which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tools for visualizing the plots.
			'TkAgg' requires the package python3-tk installed in order to run.
		save_plot : bool
			If True the generated plot is save to the defined path.

		"""
		temp_plot = plotter()
		for i in range(len(fds1.dataset)):
			try:
				temp_plot.fomplot(i,fds1, fom = 'ioff', voltages = curves[i][:,0], currents = curves[i][:,1], parameter = parameter[i],vg_ext = vg_ext,backend= backend, save_plot = save_plot)
			except (TypeError, ValueError):
				pass

class ion_ext(_extractor):
	"""
	Child class of _extractor that obtains the :math:`I_{ON}` figure of merit from a semiconductor's IV curve.

	"""

	def extraction(self,fds1,vg_ext = None, vth = None):
		"""
		Methods
		-------
		extraction(fds1,vg_ext = None, vth = None)
			Class method that extracts :math:`I_{ON}` of a semiconductor's IV curve.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for the extraction of any FoM.
		vg_ext : float
			Gate voltage value used to calculate IOFF at.
		vth : array_like
			Array of vth extracted values used for obtaining ION (using the vth-dependant formula).

		"""

		if(vth is not None):
			length = len(fds1.dataset)
			parameter = []
			curves = []
			for i in range(length):
				try:
					x_interp,y_interp = interpol(fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:],strategy=fds1.interpolation, n = 1000,s = 0)
					single_curve = np.column_stack((x_interp, y_interp))				
					ion_index=find_closest(x_interp, vth[i]+fds1.drain_bias_value)	
					parameter.append(y_interp[ion_index][0])
					curves.append(single_curve)
				except (TypeError, ValueError):
					parameter.append(np.nan)		
					curves.append(np.nan)
		elif(vg_ext is not None) and (vth is None):
			length = len(fds1.dataset)
			parameter = []
			curves = []
			for i in range(length):
				try:
					x_interp,y_interp = interpol(fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:],strategy=fds1.interpolation, n = 1000,s = 0)
					single_curve = np.column_stack((x_interp, y_interp))				
					ion_index=find_closest(x_interp, vg_ext)	
					parameter.append(y_interp[ion_index])
					curves.append(single_curve)
				except (TypeError, ValueError):
					parameter.append(np.nan)		
					curves.append(np.nan)
		else:
			print('No criteria for the extracion of ION has been defined')			
				
		return parameter, curves
#----------------------------------------------------------------------------------------------------------------	
	def save_results_to_file(self,path, parameter,fds):
		"""
		save_results_to_file(path, parameter)
			Class method that saves the extracted :math:`I_{ON}` values.

		Parameters
		----------
		path : str
			Defines the path where the extracted results are saved to a file.
		parameter : array_like
			Array of extracted FoM values to be saved into the file.

		"""
		length = len(parameter)
		f = open(str(path), "w+")
		f.write("#Sim_id\t#Ion[A]\n")
		# f.write("##############################\n###")
		for i in range(len(parameter)):
			if (str(type(parameter)) is not '<class \'numpy.ndarray\'>'):
				parameter = np.array(parameter)
			arr = np.array2string(parameter[i], separator="\t")
			directorio = fds.path_subdirs[i].split(sep='/')
			subdir = directorio[5]
			f.write("{0}\t{1}\n".format(subdir, arr))			
		f.close()
#----------------------------------------------------------------------------------------------------------------	
	def plot(self, fds1 , parameter = None, curves = None, parameter_vth = None, vg_ext = None,backend= None, save_plot = None):
		"""
		plot(fds1, parameter = None, method = None,cc_crit = None, curves = None, save = None, A=None, B=None)
			Class method that plots the extracted :math:`I_{ON}` values.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for generating the plot of any FoM.
		parameter : array_like
			Array of extracted FoM values to be plotted.
		curves : array_like
			Array of data containing the IV curves.
		parameter_vth : array_like
			Array of extracted vth values, as a method to obtain ION depends on them.		
		vg_ext : float
			Gate voltage value used to calculate IOFF at.
		backend : str
			String containing the name of the backend chosen to either plot or save the plots. The backends available are:
			'Agg', which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tools for visualizing the plots.
			'TkAgg' requires the package python3-tk installed in order to run.			
		save_plot : bool
			If True the generated plot is save to the defined path.

		"""		
		temp_plot = plotter()
		for i in range(len(fds1.dataset)):
			try:
				if(vg_ext is not None) and (type(vg_ext) is float):
					temp_plot.fomplot(i,fds1, fom = 'ion', voltages = curves[i][:,0], currents = curves[i][:,1], parameter = parameter[i], vg_ext =vg_ext,backend= backend, save_plot = save_plot)
				else:
					temp_plot.fomplot(i,fds1, fom = 'ion', voltages = curves[i][:,0], currents = curves[i][:,1], parameter = parameter[i], parameter_vth = parameter_vth[i],backend= backend, save_plot = save_plot)
			except (TypeError, ValueError):
				pass

class ss_ext(_extractor):
#----------------------------------------------------------------------------------------------------------------	
	def extraction(self, fds1, vth = None, vg_start = None, vg_end = None):
		"""
		Methods
		-------
		extraction( fds1, vg_start = None, vg_end = None)
			Class method that extracts :math:`SS` of a semiconductor's IV curve.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for the extraction of any FoM.
		vth : array_like
			Array of vth extracted values used for obtaining SS (using the vth-dependant formula).			
		vg_start : float
			Gate voltage defining the start of the interval in which the Subthreshold Swing is extracted.
		vg_end : float
			Gate voltage defining the end of the interval in which the Subthreshold Swing is extracted.

		"""
		parameter = []
		curves = []
		vt_sd_medio = []
		for i in range(len(fds1.dataset)):
			try:
				curve = np.column_stack((fds1.dataset[i][:,0][:], fds1.dataset[i][:,1][:]))

				x_interp, y_interp = interpol(curve[:,0], curve[:,1],strategy = fds1.interpolation, n = 1000,s = 0)
				if(vg_start is None) and (type(vg_end) is float):
					index = find_closest(x_interp,vg_end)
					parameter.append((x_interp[0]-x_interp[index])*1000/(np.log10(y_interp[0])-np.log10(y_interp[index])))
				elif(vg_start is not None) and (vg_end is not None) :
					index_start = find_closest(x_interp,vg_start)
					index_end = find_closest(x_interp,vg_end)
					parameter.append((x_interp[index_start]-x_interp[index_end])*1000/(np.log10(y_interp[index_start])-np.log10(y_interp[index_end])))
				elif(vg_start is not None) and (type(vg_start) is float):
					index_start = find_closest(x_interp,vg_start)
					index = find_closest(x_interp,vth[i]/2)

					parameter.append((x_interp[index_start]-x_interp[index])*1000/(np.log10(y_interp[index_start])-np.log10(y_interp[index])))
				else:
					index = find_closest(x_interp,vth[i]/2)
					parameter.append((x_interp[0]-x_interp[index])*1000/(np.log10(y_interp[0])-np.log10(y_interp[index])))
				# print(parameter)
				curves.append(curve)
				vt_sd_medio.append(vth[i]/2)
			except (TypeError, ValueError):
				parameter.append(np.nan)
				curves.append(np.nan)
				vt_sd_medio.append(np.nan)

		return parameter, curves, vt_sd_medio
#----------------------------------------------------------------------------------------------------------------	
	def save_results_to_file(self,path, parameter,fds):
		"""
		save_results_to_file(path, parameter)
			Class method that saves the extracted :math:`SS` values.

		Parameters
		----------
		path : str
			Defines the path where the extracted results are saved to a file.
		parameter : array_like
			Array of extracted FoM values to be saved into the file.

		"""		
		length = len(parameter)
		f = open(str(path), "w+")
		f.write("#Sim_id\t#SS[mV/dec]\n")
		# f.write("##############################\n###")
		for i in range(len(parameter)):
			if (str(type(parameter)) is not '<class \'numpy.ndarray\'>'):
				parameter = np.array(parameter)
			arr = np.array2string(parameter[i], separator="\t")
			directorio = fds.path_subdirs[i].split(sep='/')
			subdir = directorio[5]
			f.write("{0}\t{1}\n".format(subdir, arr))			
		f.close()
# ----------------------------------------------------------------------------------------------------------------	
	def plot(self, fds1, parameter = None, curves = None, parameter_ss = None, vg_start = None, vg_end = None, vt_sd_medio = None,backend= None, save_plot = None):
		"""
		plot(fds1, parameter = None, method = None,cc_crit = None, curves = None, save = None, A=None, B=None)
			Class method that plots the extracted :math:`SS` values.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for generating the plot of any FoM.
		parameter : array_like
			Array of extracted FoM values to be plotted.
		curves : array_like
			Array of data containing the IV curves.
		parameter_ss : array_like
			Array of extracted ss values.		
		vg_start : float
			Gate voltage defining the start of the interval in which the Subthreshold Swing is extracted.
		vg_end : float
			Gate voltage defining the end of the interval in which the Subthreshold Swing is extracted.
		vg_sd_medio : float
			Value in the middle of the interval between zero and vth extracted with the SD method.
			It is used only for defining a limit in the plot.
		backend : str
			String containing the name of the backend chosen to either plot or save the plots. The backends available are:
			'Agg', which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tools for visualizing the plots.
			'TkAgg' requires the package python3-tk installed in order to run.			
		save_plot : bool
			If True the generated plot is save to the defined path.

		"""
		temp_plot = plotter()
		for i in range(len(fds1.dataset)):
			try:
				temp_plot.fomplot(i,fds1, fom = 'ss', voltages = curves[i][:,0], currents = curves[i][:,1], parameter = parameter[i],vg_start = vg_start, vg_end = vg_end, vt_sd_medio = vt_sd_medio[i] ,backend= backend,save_plot = save_plot)
			except (TypeError, ValueError):
					pass

class dibl_ext(_extractor):

	def extraction(self,fds1,fds2,method = None, cc_criteria = None):
		"""
		Methods
		-------
		extraction(fds1,fds2,method = None)
			Class method that extracts the :math:`DIBL` of two given semiconductor's IV curves.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for the extraction of any FoM.
		fds2 : FoMpy Dataset
			additional structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for generating the plot of the calibration and the DIBL.
		method : str
			Keyword indicating the desired method of extraction of the FoMs. The list of available methods includes:
			'SD', 'CC', 'TD' and 'LE'. If method is not defined the 'SD' is selected by default.

		"""

		temp_vth = vth_ext()
		if(method == None):
			method = 'SD'
		if(method is not 'LE'):
			# fds1.drain_bias_label = 'High'
			vth_high, curve_high= temp_vth.extraction(fds1, method, cc_criteria)
			vth_low, curve_low= temp_vth.extraction(fds2, method, cc_criteria)
		else:
			# fds1.drain_bias_label = 'High'
			vth_high, curve_high, _, _= temp_vth.extraction(fds1, method, cc_criteria)
			vth_low, curve_low, _, _= temp_vth.extraction(fds2, method, cc_criteria)

		dibl = []
		corriente_low_arr = []
		for i in range(len(fds1.dataset)):
			try:
				vth_index_low = find_closest(curve_low[i][:,0],vth_low[i])
				corriente_low = curve_low[i][vth_index_low,1]

				# print(vth_high)
				# print(vth_low)
				# print(fds1.drain_bias_value)
				# print(fds2.drain_bias_value)
				# print(vth_high)
				# print(vth_low)
				dibl.append((-(vth_low[i] - vth_high[i])/(fds1.drain_bias_value-fds2.drain_bias_value)*1000))
				corriente_low_arr.append(corriente_low)
			except (TypeError, ValueError):
				dibl.append(np.nan)
				corriente_low_arr.append(np.nan)

		return dibl,curve_high, curve_low, vth_high, vth_low, corriente_low_arr

#----------------------------------------------------------------------------------------------------------------	
	def save_results_to_file(self,path, parameter, fds):
		"""
		save_results_to_file(path, parameter)
			Class method that saves the extracted :math:`DIBL` values.

		Parameters
		----------
		path : str
			Defines the path where the extracted results are saved to a file.
		parameter : array_like
			Array of extracted FoM values to be saved into the file.

		"""		
		length = len(parameter)
		f = open(str(path), "w+")
		f.write("#Simulation id\t#Parameter\n")
		f.write("##############################\n###")
		for i in range(length):
			if (str(type(parameter)) is not '<class \'numpy.ndarray\'>'):
				parameter = np.array(parameter)
			arr = np.array2string(parameter[i], separator="\t")
			f.write("{0}\t{1}\n".format(i, arr))			
		f.close()
#----------------------------------------------------------------------------------------------------------------	
	def plot(self,fds1 , curve_high = None,curve_low = None, parameter_vt_high = None, parameter_vt_low = None,corriente_low = None,backend= None, save_plot = None):
		"""
		plot(fds1, parameter = None, method = None,cc_crit = None, curves = None, save = None, A=None, B=None)
			Class method that plots the extracted :math:`DIBL` values.

		Parameters
		----------
		fds1 : FoMpy Dataset
			Structure of data containing the most important parameters of a semiconductor's IV curve.
			Needed for generating the plot of any FoM.
		curve_high : array_like
			Array of data containing the IV curves at high drain bias.
		curve_low : array_like
			Array of data containing the IV curves at low drain bias.			
		parameter_vt_high : float
			Voltage value of high drain bias.
		parameter_vt_low : float
			Voltage value of low drain bias.
		corriente_low : float
			Current at the vth value extracted for the curve at low drain bias.
		backend : str
			String containing the name of the backend chosen to either plot or save the plots. The backends available are:
			'Agg', which only works whenever saving plots to files (non-GUI) and 'TkAgg' a GUI tools for visualizing the plots.
			'TkAgg' requires the package python3-tk installed in order to run.
		save_plot : bool
			If True the generated plot is save to the defined path.

		"""
		temp_plot = plotter()
		for i in range(len(fds1.dataset)):
			try:	
			# print(i)
				temp_plot.fomplot(i,fds1, fom = 'dibl', curve_high = curve_high[i], curve_low = curve_low[i], vth_high = parameter_vt_high[i], vth_low = parameter_vt_low[i], corriente_low = corriente_low[i],backend= backend, save_plot = save_plot)
			except (TypeError, ValueError):
				pass

#----------------------------------------------------------------------------------------------------------------	

def interpol(x = None ,y = None, n = None, strategy = None, d = None, s = None):
	"""
	Wrapper function for interpolating imported data from a semiconductor's IV curve.

	Parameters
	----------
	x : array_like, shape (n,)
		1-d array containing values of the independent variable.
	y : array_like
		Array containing values of the dependent variable.
		It can have arbitrary number of dimensions, but the length along axis
		must match the length of x. Values must be finite.
	strategy : str
		Keyword for defining the selected interpolation method: The list of available methods includes:
		'akima', 'pchip' and 'linear'.
	d : int
		Degree of the smoothing spline. Must be <= 5.
		Default is k=3, a cubic spline.
	s : float
		Positive smoothing factor used to choose the number of knots. 
	
	"""

	if(strategy is None) or (strategy is 'cubic_spline'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.spline_interpol(x, y, n, d, s)

	elif(strategy is 'akima'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.akima_interpol(x, y, n)

	elif(strategy is 'pchip'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.pchip_interpol(x, y, n)

	elif(strategy is 'linear'):
		temp_interp = interpolator()
		x_interp, y_interp = temp_interp.lin_interpol(x, y, n)

	return x_interp, y_interp



import setuptools
from distutils.command.install_data import install_data
import os, re


#####VERSION CONTROL#####

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

#REQUIRES = []
#with open('requirements.txt') as f:
#    for line in f:
#        line, _, _ = line.partition('#')
#        line = line.strip()
#        if ';' in line:
#            requirement, _, specifier = line.partition(';')
#            for_specifier = EXTRAS.setdefault(':{}'.format(specifier), [])
#            for_specifier.append(requirement)
#        else:
#           REQUIRES.append(line)


setuptools.setup(
    name='mlfompy',
    #cmdclass={'install_data': custom_install},
    version=find_version("src", "mlfompy", "__init__.py"),
    description='MLFoMPy is an effective tool that extracts the main figures of merit (FoM) of a semiconductors IV curve',
    author='',
    author_email='',
    url='https://gitlab.citius.usc.es/modev/mlfompy',
    setup_requires=['setuptools','numpy'],
    install_requires=['scipy','pytest','probscale', 'matplotlib'],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    classifiers=[
        'Topic :: Utilities',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
    ],
)
# with open('README.rst', 'r') as fh:
#     long_description = fh.read()


# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools





# -*- coding: utf-8 -*-
"""
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
aux.py 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This module includes several auxiliar routines used during the usage of FoMpy.

Example
-------
In order find the closest index between a given value and a array of values, the function 
find_closest is used. For example during the extraction of IOFF::

	voltage_index = find_closest(x_interp, vg_ext)
	parameter.append(y_interp[voltage_index])

In order calculate the discrete derivate of a given curve the function get_diff is used as follows (second derivative method
extraction)::

	d2 = get_diff(curve, order = 2)
	lower_bound=find_closest(d2[:,0],10*d2[1,0])
	higher_bound=find_closest(d2[2:,0],0.8*d2[-1,0])						
	warning_interval_limit_low = int(np.round(lower_bound*1.5))
	warning_interval_limit_high = int(np.round(higher_bound*0.8))
	vt_temp = np.argmax(d2[lower_bound:higher_bound,1])

where these the maximum value of the second derivative is found leaving out values outside a defined interval where the derivatives
can be too noisy.

Finally, the function checkPath is used in FoMpy to check if a directory exists and if it does not it creates it when saving a plot or results.

"""

import numpy as np
from abc import ABCMeta, abstractmethod
import os

def find_closest(A, target):
	"""
	Function that finds the closest item inside an array A to a target value.
	It returns the index of A with the closest value to the target (:math:`A[idx]~~{\\approx}~~target`).

	Parameters
	-------
	A : array_like
		Array containing all the values to be compared. It usually is
		an array of currents of a FoMpy dataset.
	target : float
		Target value to be found inside the array A.
	Returns
	-------
	idx : int
		Index of the item contained in the array A with a value closest to the target.
	"""
	idx = A.searchsorted(target) #A must be sorted
	idx = np.clip(idx, 1, len(A)-1)
	left = A[idx-1]
	right = A[idx]
	idx -= target - left < right - target
	return idx
#----------------------------------------------------------------------------------------------------------------
def get_diff(arr, order, type = 'central'):
	"""
	Function for calculating the n-order derivative of a discrete array

	Parameters
	-------
	order : int
		Order indicating the type of derivative to compute
	type: str
		The type of derivate can either be calculated in a forward way, 
		central or backward, depending on the evaluation coordinates
		of the derivative function (see https://en.wikipedia.org/wiki/Finite_difference)
	"""

	x = arr[:,0]
	y = arr[:,1]

	dx = np.gradient(x)
	dy = np.gradient(y)
	d1 = dy/dx
	d2y = np.gradient(d1)
	d2 = d2y/dx
	d3y = np.gradient(d2y)
	d3 = d3y/dx

	if order == 1:
		gm_c = np.column_stack((x, d1))
		return(gm_c)

	if order == 2:
		gm2_c = np.column_stack((x, d2))
		return(gm2_c)

	if order == 3:
		gm3_c = np.column_stack((x, d3))
		return(gm3_c)
#----------------------------------------------------------------------------------------------------------------
def checkPath(cwd):
	"""
	Function that checks if a given path exists, and if it doesn't, it creates it.

	Parameters
	-------
	cwd : str
		Path that the user wishes to check if it exists.
	"""
	try:
		os.makedirs(cwd)
	except OSError:
		if not os.path.isdir(cwd):
			raise
